"use strict";

const EXPORTED_SYMBOLS = ["branding"];

const { classes: Cc, interfaces: Ci, utils: Cu, results: Cr } = Components;
const GLOBAL = this;

const SVC_COMPONENT_ID = "ru.yandex.custombar.branding",
      SVC_NAME = "package",
      SVC_PKG_UPD_TOPIC = "package updated";


const brandingService = {
    
    findFile: function PartnerPackSvc_findFile(filePath) {
        return branding.brandPackage.findFile(filePath);
    },
    
    
    getXMLDocument: function PartnerPackSvc_getXMLDocument(docPath, privilegedParser) {
        try {
            return branding.brandPackage.getXMLDocument(docPath, privilegedParser);
        }
        catch (e) {
            branding._logger.error(strutils.formatString("Could not return XML-document for file '%1'. \nError: %2", [docPath, strutils.formatError(e)]));
            throw e;
        }
    },
    
    getBrandID: function PartnerPackSvc_getBrandID() {
        return branding.productInfo.BrandID;
    },
    
    getPackageDate: function PartnerPackSvc_getPackageDate() {
        return branding.brandingDate;
    },
    
    
    resolvePath: function PartnerPackSvc_resolvePath(path, base) {
        return branding.brandPackage.resolvePath(path, base);
    },
    
    
    expandBrandTemplates: function PartnerPackSvc_expandBrandTemplates(templateString, params) {
        return branding.expandBrandTemplates(templateString, params);
    },
    
    
    expandBrandTemplatesEscape: function PartnerPackSvc_expandBrandTemplatesEscape(templateString, params) {
        return branding.expandBrandTemplatesEscape(templateString, params);
    },
    
    
    makeURLTemplate: function PartnerPackSvc_makeURLTemplate(templateString) {
        return branding.makeURLTemplate(templateString);
    },
    
    
    getYandexFeatureState: function PartnerPackSvc_getYandexFeatureState(aFeatureName) {
        return branding.getYandexFeatureState(aFeatureName);
    }
};

const branding = {
    init: function PartnerPack_init(barApp) {
        this._logger = barApp.getLogger("Branding");
        this._barApp = barApp;
        barApp.core.Lib.sysutils.copyProperties(barApp.core.Lib, GLOBAL);
        
        this._loadPackage();
        
        packageDownloader.init(barApp);
        xbDataProvider.init(barApp);
        
        barApp.NativeComponents.registerService(SVC_COMPONENT_ID, SVC_NAME, brandingService);
        
        Cc["@mozilla.org/network/protocol;1?name=bar"].getService().wrappedJSObject
            .addDataProvider(this._BarProtocolHandler);
        
        Cc["@mozilla.org/network/protocol;1?name=" + barApp.name]
            .getService(Ci.nsIProtocolHandler)
            .wrappedJSObject
            .addDataProvider(this._AppProtocolHandler);
        
        try {
            if (this._barApp.addonManager.info.addonVersionChanged)
                this._checkBPAndReplaceIfSame();
        }
        catch (e) {
            this._logger.error("Failed replacing the same BP with internal version. \n" + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
        
        this._trySetAddonDescription();
        this._loadDefaultBrowserPreferences();
        if (!this.Installer.checkLicenseAccepted())
            throw new Error("License agreement rejected");
        
        let addonManagerInfo = barApp.addonManager.info;
        if (addonManagerInfo.addonVersionChanged) {
            const ObserverService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
            ObserverService.addObserver(this, "sessionstore-windows-restored", false);
            
            if (!addonManagerInfo.isFreshAddonInstall && addonManagerInfo.addonUpgraded)
                this.Installer._onAddonUpdate();
        }
        
        packageDownloader.sheduleUpdates(this.productInfo.BrandingURL);
    },
    
    observe: function PartnerPack_observe(aSubject, aTopic, aData) {
        switch (aTopic) {
            
            case "sessionstore-windows-restored":
                this.Installer.showWelcomePageOnStartup();
                break;
            
            default:
                return;
        }
        
        const ObserverService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        ObserverService.removeObserver(this, aTopic, false);
    },
    
    newPackageReady: function PartnerPack_newPackageReady(packageFile, responseDate) {
        let tmpDir = this._barApp.directories.vendorDir;
        tmpDir.append("tmp");
        tmpDir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, parseInt("0755", 8));
        
        let newProductInfo;
        try {
            fileutils.extractZipArchive(packageFile, tmpDir);
            [, newProductInfo] = this._validateBrandPkg(tmpDir, true);
        }
        catch (e) {
            this._logger.error("Invalid package archive. \n" + strutils.formatError(e));
            fileutils.removeFileSafe(tmpDir);
            return;
        }
        
        this.updateFrom(tmpDir.clone(), newProductInfo, responseDate);
    },
    
    updateFrom: function Branding_updateFrom(newPkgDir, newProductInfo, responseDate) {
        let prevBrandMap;
        let prevBrowserConf;
        try {
            prevBrandMap = this._brandTemplateMap;
            prevBrowserConf = this.browserConf;
        }
        catch (e) {
            this._logger.error("Could not get previous configs representation. " + e);
            this._logger.debug(e.stack);
        }
        
        try {
            fileutils.safeReplace(this._currPackageDir, newPkgDir);
        }
        catch (e) {
            this._logger.error("Could not replace old directory with a new one. \n" + strutils.formatError(e));
            return;
        }
        finally {
            fileutils.removeFileSafe(newPkgDir);
        }
        this._updatePackage(newProductInfo, responseDate, prevBrandMap, prevBrowserConf);
    },
    
    finalize: function PartnerPack_finalize() {
        Cc["@mozilla.org/network/protocol;1?name=" + this._barApp.name]
            .getService(Ci.nsIProtocolHandler)
            .wrappedJSObject
            .removeDataProvider(this._AppProtocolHandler);
        
        Cc["@mozilla.org/network/protocol;1?name=bar"].getService().wrappedJSObject
            .removeDataProvider(this._BarProtocolHandler);
        
        xbDataProvider.finalize();
        packageDownloader.stopUpdates();
        if (this._package) {
            this._barApp.NativeComponents.unregisterService(SVC_COMPONENT_ID, SVC_NAME);
            this._package.finalize();
            this._package = null;
        }
        this._barApp = null;
    },
    
    get brandPackage() {
        return this._package;
    },
    
    get brandingDate() {
        return this._brandingDate;
    },
    
    get barless() {
        if (this._barless === null) {
            try {
                this._barless = strutils.xmlAttrToBool(this.productInfo.Barless && this.productInfo.Barless.enabled);
            } catch (e) {
                this._barless = false;
            }
        }
        return this._barless;
    },
    getYandexFeatureState: new function() {
        let cache = {};
        return function PartnerPack_getYandexFeatureState(aFeatureName) {
            if (aFeatureName in cache)
                return cache[aFeatureName];
            return cache[aFeatureName] = strutils.xmlAttrToBool( this.productInfo.YandexFeatures[aFeatureName] );
        }
    },
    
    expandBrandTemplatesEscape: function PartnerPack_expandBrandTemplatesEscape(aTemplateStr, aParams) {
        return this.expandBrandTemplates(aTemplateStr, aParams, true);
    },
    
    
    expandBrandTemplates: function PartnerPack_expandBrandTemplates(aTemplateString, aParams, aEncodeParams) {
        let result = "" + aTemplateString;
        
        if (!/\{/.test(result))
            return result;
        
        function encode(str)
            aEncodeParams ? encodeURIComponent(str) : str;
        
        let self = this;
        
        function replacer(aMatch) {
            let match = aMatch.replace(/[{}]/g, "");
            if (aParams && aParams.hasOwnProperty(match))
                return encode(aParams[match]);
            
            if (match == "brandID")
                return encode(self.productInfo.BrandID);
            
            if (/clid(.+)/.test(match)) {
                let clidData = self._vendorData[match];
                if (!(clidData && clidData.clidAndVid))
                  return "";
                
                return encode(clidData.clidAndVid);
            }
            
            const namesRe = /^(product[12]|vendor)\.(\w+)$/;
            let matchNames = match.match(namesRe);
            if (matchNames) {
                try {
                    switch (matchNames[1]) {
                        case "product1":
                            return encode(self.productInfo.ProductName1[matchNames[2]] || "");
                        case "product2":
                            return encode(self.productInfo.ProductName2[matchNames[2]] || "");
                        case "vendor":
                            return encode(self.productInfo.VendorName[matchNames[2]] || "");
                    }
                }
                catch (e) {
                    self._logger.error("Could not expand branding templates. \n" + strutils.formatError(e));
                    return "";
                }
            }
            
            return aMatch;
        }
        
        return result.replace(/\{[^\{\}]+\}/g, replacer);
    },
    
    makeURLTemplate: function PartnerPack_makeURLTemplate(aTemplateStr) {
    
        let template = $wc.$uri.$Template(aTemplateStr);
        let tplMap = this._brandTemplateMap;
        
        return function PartnerPack_URITemplate_instance(moreKeys) {
            return template(sysutils.mergeObj(tplMap, moreKeys));
        };
    },
    
    get productInfo() {
        if (this._productInfo === undefined) {
            try {
                this._productInfo = this._getBPProductInfo(this._package);
            } catch (e) {
                this._logger.error("Could not read file '/about/product.xml' \n" + strutils.formatError(e));
                this._productInfo = null;
            }
        }
        return this._productInfo;
    },
    
    get browserConf() {
        if (this._browserConf === undefined) {
            try {
                this._browserConf = xmlutils.dom2jsObj(this._package.getXMLDocument("/browser/browserconf.xml"));
            } catch (e) {
                this._logger.error("Could not read file '/browser/browserconf.xml'");
                this._browserConf = null;
            }
        }
        return this._browserConf;
    },
    _barApp: null,
    _logger: null,
    _package: null,
    _productInfo: undefined,
    _browserConf: undefined,
    _updateUrl: undefined,
    _barless: null,
    
    _consts: {
        GCASES: ["nom", "gen", "dat", "acc", "ins", "pre", "loc", "ex1", "ex2"],
        BRAND_TS_PREF_NAME: "branding.lastupdate",
        PKG_DIR_NAME: "branding"
    },
    
    get _brandingDate() {
        let brandTimestamp = Number( this._barApp.preferences.get(this._consts.BRAND_TS_PREF_NAME, undefined) );
        let brandingDate =  brandTimestamp ? new Date(brandTimestamp * 1000) : this._barApp.core.buildDate;
        if (!brandTimestamp)
            this._barApp.preferences.overwrite(this._consts.BRAND_TS_PREF_NAME, brandingDate.getTime() / 1000);
        
        return brandingDate;
    },
    
    set _brandingDate(newDate) {
        this._barApp.preferences.overwrite(this._consts.BRAND_TS_PREF_NAME, newDate.getTime() / 1000);
    },
    _beforeBrandReplacement: function PartnerPack__beforeBrandReplacement() {
        this._brandTemplateMap;
        this.browserConf;
    },
    
    _afterBrandReplacement: function PartnerPack__afterBrandReplacement() {
        let [, newProductInfo] = this._validateBrandPkg(this._currPackageDir, true);
        this._updatePackage(newProductInfo, new Date(), this._brandTemplateMap, this.browserConf);
    },
    
    get _currPackageDir() {
        let currPkgDir = this._barApp.directories.vendorDir;
        currPkgDir.append(this._consts.PKG_DIR_NAME);
        return currPkgDir;
    },
    
    _loadPackage: function PartnerPack__loadPackage() {
        let package_;
        let productInfo;
        
        try {
            let currPkgDir = this._currPackageDir;
            if (!currPkgDir.exists())
                this._barApp.addonFS.copySource("$custombar/branding",
                                                this._barApp.directories.vendorDir,
                                                this._consts.PKG_DIR_NAME);
            [package_, productInfo] = this._validateBrandPkg(currPkgDir);
        }
        catch (e) {
            this._logger.error("Existing branding package is invalid! Will try internal version. \n" + strutils.formatError(e));
            this._logger.debug(e.stack);
            fileutils.removeFileSafe(this._currPackageDir);
            this._barApp.addonFS.copySource("$custombar/branding", this._barApp.directories.vendorDir, this._consts.PKG_DIR_NAME);
            [package_, productInfo] = this._validateBrandPkg(this._currPackageDir);
            
            this._brandingDate = this._barApp.core.buildDate;
        }
        
        this._package = package_;
        this._productInfo = productInfo;
    },
    _checkBPAndReplaceIfSame: function PartnerPack__checkBPAndReplaceIfSame() {
        // If there's no BP update timestamp available, build date is used. Thus these dates will be equal (BAR-9415).
        if (this._brandingDate > this._barApp.core.buildDate) {
            this._logger.debug(strutils.formatString("Current BP is newer (%1) than my build date (%2)",
                                                     [this._brandingDate, this._barApp.core.buildDate]));
            return;
        }
        
        let currentProductInfo = this._productInfo;
        
        let internalPkgDir = this._barApp.directories.vendorDir;
        internalPkgDir.append("tmp");
        if (internalPkgDir.exists())
            internalPkgDir.remove(true);
        this._barApp.addonFS.copySource("$custombar/branding", this._barApp.directories.vendorDir, "tmp");
        
        let internalProductDoc;
        let internalTmpPkg = new this._barApp.BarPlatform.FilePackage(internalPkgDir);
        try {
            internalProductDoc = internalTmpPkg.getXMLDocument("/about/product.xml");
        }
        finally {
            internalTmpPkg.finalize();
            if (!internalProductDoc)
                fileutils.removeFileSafe(internalPkgDir);
        }
        
        let internalBUrl = xmlutils.queryXMLDoc("string(/Product/BrandingURL)", internalProductDoc);
        if (currentProductInfo.BrandingURL != internalBUrl) {
            fileutils.removeFileSafe(internalPkgDir);
            return;
        }
        this._logger.debug("Replacing BP with the same updateURL with internal one");
        
        let [, newProductInfo] = this._validateBrandPkg(internalPkgDir, true);
        this.updateFrom(internalPkgDir, newProductInfo, this._barApp.core.buildDate);
    },
    
    
    _trySetAddonDescription: function PartnerPack__trySetAddonDescription() {
        let productInfo = this.productInfo;
        let prefs = this._barApp.preferences;
        
        try {
            prefs.set("name", productInfo.ProductName1.nom);
            prefs.set("description", productInfo.ProductDescription.fx);
            prefs.set("creator", productInfo.VendorName.nom);
            
            let homepageURL = productInfo.HomePage && productInfo.HomePage.toString() || "";
            prefs.set("homepageURL", homepageURL);
        }
        catch (e) {
            this._logger.error("Could not set branded addon description. " + strutils.formatError(e));
        }
        
        if (productInfo.BrandID.toString() == "yandex")
            return;
        let cacheEnabledPrefName = "getAddons.cache.enabled";
        if (prefs.get(cacheEnabledPrefName, null) === false &&
            !this._barApp.addonManager.info.addonVersionChanged)
            return;
        
        prefs.set(cacheEnabledPrefName, false);
        if (!sysutils.platformInfo.browser.version.isGreaterThan("3.6"))
            return;
        try {
            let addonsDBFile = Cc["@mozilla.org/file/directory_service;1"]
                                   .getService(Ci.nsIProperties)
                                   .get("ProfD", Ci.nsIFile);
            addonsDBFile.append("addons.sqlite");
            
            if (addonsDBFile.exists() && addonsDBFile.isFile()) {
                let addonsDB = new misc.Database(addonsDBFile);
                let addonId = this._barApp.addonManager.addonId;
                addonsDB.execQueryAsync("DELETE FROM addon WHERE id = :id", {id: addonId});
                addonsDB.close();
                this._logger.debug("Cleared add-on info from browser database.");
            } else {
                this._logger.warn("Can't find browser add-ons database file.");
            }
        }
        catch (e) {
            this._logger.error("Could not reset add-on description. " + strutils.formatError(e));
        }
    },
    
    _validateBrandPkg: function PartnerPack__validateBrandPkg(packageDir, autoFinalizePkg) {
        let package_ = new this._barApp.BarPlatform.FilePackage(packageDir);
        let productInfo = xmlutils.dom2jsObj(package_.getXMLDocument("/about/product.xml"));
        
        this._logger.trace("Checking product info:\n" + sysutils.dump(productInfo, 10));
        
        if (!productInfo.BrandID || (productInfo.BrandID == ""))
            throw new Error("No brand ID");
        if (!productInfo.ProductName1.nom)
            throw new Error("No product name");
        try {
            let updateURL = productInfo.BrandingURL;
            netutils.newURI(updateURL, null, null);
        }
        catch (e) {
            throw new Error("Branding update URL unavailable. \n" + strutils.formatError(e));
        }
        
        try {
            let fxProductInfo = xmlutils.dom2jsObj(package_.getXMLDocument("/fx/about/product.xml"));
            for (let featureName in fxProductInfo.YandexFeatures) {
                if (["toString", "childElements"].indexOf(featureName) == -1)
                    productInfo.YandexFeatures[featureName] = fxProductInfo.YandexFeatures[featureName];
            }
        } catch (e) {}
        
        if (autoFinalizePkg) {
            try {
                package_.finalize();
            }
            catch (e) {
                this._logger.error("Error finalizing temporary package instance. \n" + strutils.formatError(e));
            }
            package_ = null;
        }
        
        return [package_, productInfo];
    },
    
    _updatePackage: function PartnerPack__updatePackage(newProductInfo, responseDate, prevBrandMap, prevBrowserConf) {
        this._package.finalize();
        this._browserConf = undefined;
        this._brandTemplateMap = undefined;
        
        this._package = new this._barApp.BarPlatform.FilePackage(this._currPackageDir);
        this._productInfo = newProductInfo;
        
        packageDownloader.sheduleUpdates(newProductInfo.BrandingURL);
        
        this._brandingDate = responseDate;
        this._barApp.preferences.set(this._consts.BRAND_TS_PREF_NAME, Math.round(responseDate.getTime() / 1000));
        this._logger.info("Package updated");
        
        xbDataProvider.packageUpdated();
        this._applyPartnerSettings(prevBrandMap, prevBrowserConf);
        this._barApp.NativeComponents.notifyServiceUsers(SVC_COMPONENT_ID, SVC_NAME, SVC_PKG_UPD_TOPIC, null);
    },
    _applyPartnerSettings: function PartnerPack__applyPartnerSettings(prevBrandMap, prevBrowserConf) {
        try {
            this.Installer._onPackageUpdate(prevBrandMap, prevBrowserConf);
        }
        catch (e) {
            this._logger.error("Could not apply partner package settings. \n" + strutils.formatError(e));
        }
    },
    get _vendorData() {
        let oldCore = this._barApp.core.oldCore;
        let vendorData = oldCore ? sysutils.copyObj(oldCore.vendorData) : {};
        
        delete this._vendorData;
        return this._vendorData = vendorData;
    },
    
    get _brandTemplateMap() {
        if (!this.__brandMap)
            this.__brandMap = this._makeBrandTemplateMap(this._package);
        return this.__brandMap;
    },
    
    set _brandTemplateMap(val) {
        this.__brandMap = val;
    },
    _makeBrandTemplateMap: function PartnerPack__makeBrandTemplateMap(brandPackage, aExtraKeys, aEncodeParams) {
        let productInfo = this._getBPProductInfo(brandPackage || this._package);
        
        function encode(str)
            aEncodeParams ? encodeURIComponent(str) : str;
        
        let result = {};
        
        result.brandID = encode(productInfo.BrandID);
        
        for each (let [, gCase] in Iterator(this._consts.GCASES)) {
            result["product1." + gCase] = encode(productInfo.ProductName1[gCase] || "");
            result["product2." + gCase] = encode(productInfo.ProductName2[gCase] || "");
            result["vendor." + gCase] = encode(productInfo.VendorName[gCase] || "");
        }
        
        for each (let [clidName, clidData] in Iterator(this._vendorData))
            result[clidName] = encode(clidData ? clidData.clidAndVid : "");
        
        if (aExtraKeys) {
            for each (let [tplName, tplValue] in Iterator(aExtraKeys))
                result[tplName] = encode(tplValue);
        }
        
        return result;
    },
    
    _getBPProductInfo: function Branding__getBPProductInfo(brandPackage) {
        let productInfo = xmlutils.dom2jsObj(brandPackage.getXMLDocument("/about/product.xml"));
        try {
            let fxProductInfo = xmlutils.dom2jsObj(brandPackage.getXMLDocument("/fx/about/product.xml"));
            for (let featureName in fxProductInfo.YandexFeatures) {
                if (["toString", "childElements"].indexOf(featureName) == -1)
                    productInfo.YandexFeatures[featureName] = fxProductInfo.YandexFeatures[featureName];
            }
        } catch (e) {}
        return productInfo;
    },
    
    _getBPBrowserConf: function Branding__getBPBrowserConf(brandPackage) {
        return xmlutils.dom2jsObj(brandPackage.getXMLDocument("/browser/browserconf.xml"));
    },
    
    _BarProtocolHandler: {
        INDEX_FILE_NAME: "welcome.html",
        
        get WP_XB_PATH() {
            let wpPath = "/welcome/fx/";
            let brandPackage = branding.brandPackage;
            let wpURL = brandPackage.findFile(wpPath + this.INDEX_FILE_NAME) && brandPackage.resolvePath(wpPath);
            
            delete this.WP_XB_PATH;
            return this.WP_XB_PATH = wpURL;
        },
        
        newURI: function BarPH_WP_newURI(aSpec, aOriginalCharset, aBaseURI, aSimpleURI) {
            if (!aBaseURI && this.canHandlePath(aSimpleURI.path))
                return aSimpleURI;
            if (aBaseURI && this.canHandlePath(aBaseURI.path))
                return this._makeWPURI(aSpec, aOriginalCharset, null);
            
            return null;
        },
        
        newChannel: function BarPH_WP_newChannel(aURI) {
            if (!this.canHandlePath(aURI.path))
                return null;
            let baseURI = this._makeWPURI("", null, null);
            let uri = this._makeWPURI(this.INDEX_FILE_NAME, null, baseURI);
            let channel = this._ioService.newChannelFromURI(uri);
            channel.originalURI = aURI;
            return channel;
        },
        
        canHandlePath: function BarPH_WP_canHandlePath(aPath) {
            return (aPath.toLowerCase() == "welcome") && this.WP_XB_PATH;
        },
        
        _makeWPURI: function BarPH_WP__makeURI(aTarget, aOriginCharset, aBaseURI) {
            return this._ioService.newURI(this.WP_XB_PATH + aTarget, aOriginCharset, aBaseURI);
        },
        
        get _ioService() {
            delete this._ioService;
            return this._ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        }
    },
    
    _AppProtocolHandler: {
        
        newURI: function AppPH_newURI(aSpec, aOriginalCharset, aBaseURI, simpleURI) {
            if (!aSpec)
                return null;
            
            let spec = aBaseURI ? aBaseURI.resolve(aSpec) : aSpec;
            let resourcePath = spec && spec.split(/:\/\/branding\//)[1];
            let brandingPackageURL = resourcePath && branding.brandPackage.resolvePath(resourcePath);
            
            return misc.tryCreateFixupURI(brandingPackageURL);
        }
    },
    
    _loadDefaultBrowserPreferences: function PartnerPack__loadDefaultBrowserPreferences() {
        if (this.getYandexFeatureState("safe-browsing")) {
            let sbPrefName = this._barApp.preferencesBranch + "safebrowsing.installed";
            if (!Preferences.get(sbPrefName, false)) {
                Preferences.set(sbPrefName, true);
                
                let safeBrowsingFile = this._barApp.core.extensionPathFile;
                "defaults/dynamic-preferences/safebrowsing.js"
                    .split("/").forEach(function(s) safeBrowsingFile.append(s));
                
                this._barApp.preferences.loadFromFile(safeBrowsingFile);
            }
        }
    },
    
    get Installer() {
        const barApp = this._barApp;
        const brandPrefs = new Preferences(barApp.preferencesBranch + "branding.");
        
        const Installer = {
            checkLicenseAccepted: function Installer_checkLicenseAccepted() {
                const acceptedPrefName = barApp.name + ".license.accepted";
                let needRestartBrowser = false;
                
                if (!Preferences.get(acceptedPrefName, false)) {
                    if (!this.setupData.hiddenWizard) {
                        try {
                            let accepted = this._showLicenseWindow();
                            if (!accepted) {
                                this._disableAddonOnLicenseRefuse();
                                return false;
                            }
                        } catch (e) {
                            Cu.reportError(e);
                        }
                    }
                    
                    Preferences.set(acceptedPrefName, true);
                    if (Preferences.get(acceptedPrefName, false) === true)
                        needRestartBrowser = true;
                    try {
                        let prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService);
                        prefService.savePrefFile(null);
                    } catch(e) {
                        needRestartBrowser = false;
                        Cu.reportError(e);
                    }
                    
                    this._onAddonInstall();
                }

                try {
                    this._setActiveQS();
                } catch (e) {
                    Cu.reportError(e);
                }
                try {
                    let installInfo = barApp.addonManager.info;
                    if (installInfo.addonVersionChanged &&
                        sysutils.versionComparator.compare(installInfo.addonLastVersion, "5.2.3") <= 0) {
                        this._writeQuickSearches(false);
                        
                        let yandexString = "\u042f\u043d\u0434\u0435\u043a\u0441";
                        let selectedEngineName = Preferences.get("browser.search.selectedEngine", null);
                        let defaultEngineName = Preferences.get("browser.search.defaultenginename", null);
                        
                        if (selectedEngineName === yandexString || (!selectedEngineName && defaultEngineName === yandexString)) {
                            try {
                                let yaQSShortName = branding.brandPackage.getXMLDocument("/browser/browserconf.xml")
                                               .querySelector("Browser > QuickSearch > OpenSearchDescription[*|uniqName='yandex'] > ShortName");
                                yaQSShortName = (yaQSShortName && yaQSShortName.textContent) || "";
                                
                                if (yaQSShortName)
                                    Preferences.set("browser.search.selectedEngine", yaQSShortName);
                            } catch (e) {
                                Cu.reportError(e);
                            }
                        }
                    }
                } catch (e) {
                    Cu.reportError(e);
                }
                
                if (needRestartBrowser) {
                    this._logger.debug("Restart browser.");
                    const appStartup = Cc["@mozilla.org/toolkit/app-startup;1"].getService(Ci.nsIAppStartup);
                    appStartup.quit(appStartup.eForceQuit | appStartup.eRestart);
                    return null;
                }
                let selectedEngineName = Preferences.get("browser.search.selectedEngine", null);
                let defaultEngineName = Preferences.get("browser.search.defaultenginename", null);
                if (!selectedEngineName && defaultEngineName)
                    Preferences.set("browser.search.selectedEngine", defaultEngineName);
                
                return true;
            },
            _setActiveQS: function Installer__setActiveQS() {
                const hackPrefName = "qsEngineNameForSelect";
                
                let qsNameToSet = brandPrefs.get(hackPrefName, null);
                if (qsNameToSet) {
                    brandPrefs.reset(hackPrefName);
                    Preferences.set("browser.search.selectedEngine", qsNameToSet);
                }
            },
            
            _setupData: null,
            
            set setupData(aValue) {
                this._setupData = null;
            },
            
            get setupData() {
                if (!this._setupData) {
                    let data = {
                        hiddenWizard: true,
                        
                        License: {
                            display: false,
                            checked: false,
                            text: ""
                        },
                        
                        HomePage: {
                            display: false,
                            checked: false,
                            text: ""
                        },
                        
                        DefaultSearch: {
                            display: false,
                            checked: false,
                            text: ""
                        }
                    };
                    
                    let setupElement;
                    
                    try {
                        let productXML = branding.brandPackage.getXMLDocument("/about/product.xml");
                        setupElement = productXML.querySelector("Product > Setup");
                    } catch (e) {
                        Cu.reportError(e);
                    }
                    
                    if (setupElement) {
                        ["License", "HomePage", "DefaultSearch"]
                        .forEach(function(aElementName) {
                            let el = setupElement.querySelector(aElementName);
                            if (el) {
                                for (let [prop, val] in Iterator(data[aElementName])) {
                                    let attrValue = el.getAttribute(prop);
                                    if (typeof val == "boolean")
                                        attrValue = (attrValue == "true");
                                    data[aElementName][prop] = attrValue;
                                }
                            }
                            
                            if (data.hiddenWizard && data[aElementName].display)
                                data.hiddenWizard = false;
                        });
                    }
                    
                    data.License.url = branding.brandPackage.resolvePath("/license/fx/license.xhtml");
                    
                    data.HomePage.title = "";
                    data.HomePage.url = "";
                    data.HomePage.force = false;
                    
                    data.LocationBarSearch = {
                        url: "",
                        urlForce: false
                    };
                    
                    try {
                        let configXML = branding.brandPackage.getXMLDocument("/browser/browserconf.xml");
                        let configHPElement = configXML.querySelector("Browser > HomePage");
                        
                        if (configHPElement) {
                            let homePageURL = configHPElement.textContent;
                            data.HomePage.url = homePageURL && branding.expandBrandTemplatesEscape(homePageURL);
                            data.HomePage.title = configHPElement.getAttribute("title") || "";
                            data.HomePage.force = configHPElement.getAttribute("force") === "true";
                        }
                        
                        let configLBSElement = configXML.querySelector("Browser > LocationBarSearch > Url");
                        if (configLBSElement) {
                            data.LocationBarSearch.url = branding.expandBrandTemplatesEscape(configLBSElement.textContent);
                            data.LocationBarSearch.urlForce = configLBSElement.getAttribute("force") === "true";
                        }
                    } catch (e) {
                        this._logger.error(e);
                        this._logger.debug(e.stack);
                    }
                    
                    let stringBundle = new barApp.appStrings.StringBundle("custombar/dialogs/license/wizard.properties");
                    if (!data.HomePage.text && data.HomePage.display) {
                        data.HomePage.text = stringBundle.get("homepage.label");
                        if (branding.getYandexFeatureState("homepage-protection"))
                            data.HomePage.text += " " + stringBundle.get("homepageProtect.label");
                    }
                    if (!data.DefaultSearch.text && data.DefaultSearch.display) {
                        data.DefaultSearch.text = stringBundle.get("defaultsearch.label");
                    }
                    
                    data.productName = branding.productInfo.ProductName1.nom;
                    
                    this._setupData = data;
                }
                
                return this._setupData;
            },
            
            getBrowserHomePage: function Installer_getBrowserHomePage() {
                const browserHPPrefName = "browser.startup.homepage";
                let currentHP;
                try {
                    currentHP = Cc["@mozilla.org/preferences-service;1"]
                                    .getService(Ci.nsIPrefService)
                                    .QueryInterface(Ci.nsIPrefBranch2)
                                    .getComplexValue(browserHPPrefName, Ci.nsIPrefLocalizedString)
                                    .data;
                } catch(e) {}

                if (!currentHP) {
                    const SBS = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService);
                    let propertiesPath = !!Ci.nsIContentSecurityPolicy ?
                                             "chrome://branding/locale/browserconfig.properties" :
                                             "resource:/browserconfig.properties";
                    let configBundle = SBS.createBundle(propertiesPath);
                    currentHP = configBundle.GetStringFromName(browserHPPrefName);
                }
                
                return currentHP;
            },
            
            setBrowserHomePage: function Installer_setBrowserHomePage(aHomePageURL) {
                let url = arguments.length ? aHomePageURL : this.setupData.HomePage.url;
                this._setBrowserHomePage(url, true);
            },
            
            get _appLocale() {
                let locale;
                const xulChromeReg = Cc["@mozilla.org/chrome/chrome-registry;1"].getService(Ci.nsIXULChromeRegistry);
                try {
                    locale = xulChromeReg.getSelectedLocale("global");
                } catch (e) {
                    this._logger.error("Can't get application locale. " + e);
                }
                
                delete this._appLocale;
                return this._appLocale = locale || "en";
            },
            
            isYandexURL: function Installer_isYandexURL(aURL) {
                return !!(aURL &&
                          ((aURL == "about:home" && this._appLocale == "ru") ||
                           /^(https?:\/\/)?(www\.)?yandex\.(ru|ua|kz|by|com|com\.tr)(\/|$)/i.test(aURL) ||
                           /^(https?:\/\/)?ya\.ru(\/|$)/i.test(aURL) ||
                           /^(https?:\/\/)?ru\.start\d+\.mozilla\.com\/firefox/.test(aURL)
                           )
                       );
            },
            isOverridableURL: function Installer_isOverridableURL(aURL) {
                return !(aURL &&
                         (this.isYandexURL(aURL) ||
                          /^(https?:\/\/)?((www|search)\.)?seznam\.cz(\/|$)/i.test(aURL) ||
                          /^(https?:\/\/)?(www\.)?bozzon\.com(\/|$)/i.test(aURL)
                         )
                       );
            },
            isCurrentQSOverridable: function Installer_isCurrentQSOverridable() {
                let selectedEngineName = Preferences.get("browser.search.selectedEngine", null) ||
                                         Preferences.get("browser.search.defaultenginename", null);
                return this.isQSOverridable(selectedEngineName);
            },
            isQSOverridable: function Installer_isQSOverridable(aQSName) {
                return [
                    "\u042f\u043d\u0434\u0435\u043a\u0441",
                    "Yandex",
                    "Seznam",
                    "Bozzon"
                ].indexOf(aQSName) == -1;
            },
            
            showWelcomePageOnStartup: function Installer_showWelcomePageOnStartup() {
                if (!branding._BarProtocolHandler.canHandlePath("welcome"))
                    return;
                
                let needShow = true;
                
                let wpPrefs = new Preferences(barApp.name + ".welcomepage.");
                
                let wpVersion = wpPrefs.get("version", "0");
                let wpVersionIntroduced = wpPrefs.get("version.introduced", "0");
                
                if (wpPrefs.get("dontshow", true)) {
                    wpPrefs.reset("dontshow");
                    needShow = false;
                }
                
                const versionComparator = Cc["@mozilla.org/xpcom/version-comparator;1"].getService(Ci.nsIVersionComparator);
                if (wpVersionIntroduced && versionComparator.compare(wpVersion, wpVersionIntroduced) < 1) {
                    needShow = false;
                } else {
                    wpVersionIntroduced = wpVersion;
                }
                
                wpPrefs.set("version.introduced", wpVersionIntroduced);
                
                if (!needShow)
                    return;
                
                new sysutils.Timer((function(aBarApp) {
                    return function Installer_showWelcomePageOnStartup_timed() {
                        aBarApp.navigate({
                            url: "bar:welcome",
                            target: "new tab"
                        });
                    }
                })(barApp), 500);
            },
            _onAddonInstall: function Installer__onAddonInstall() {
                let setupData = this.setupData;
                
                if (setupData.HomePage.checked)
                    this._setBrowserHomePage(setupData.HomePage.url, false, true);
                
                if (setupData.DefaultSearch.checked)
                    this._setLocationBarSearch(true);
                
                this._writeQuickSearches(setupData.DefaultSearch.checked);
                
                let dataForDistribution = this._getDataForDistribution();
                dataForDistribution.isHomepageChecked = setupData.HomePage.checked;
                dataForDistribution.isSearchChecked = setupData.DefaultSearch.checked;
                barApp.distribution.onInstall(dataForDistribution);
            },
            _onAddonUpdate: function Installer__onAddonUpdate() {
                barApp.distribution.onUpdate(this._getDataForDistribution());
            },
            _onPackageUpdate: function Installer__onPackageUpdate(prevBrandMap, prevBrowserConf) {
                this.setupData = null;
                let setupData = this.setupData;
                
                this._logger.debug("prevBrowserConf " + sysutils.dump(prevBrowserConf, 5));
                
                let prevBrandHP = branding.expandBrandTemplatesEscape(prevBrowserConf.HomePage, prevBrandMap);
                this._logger.debug("prevBrandHP " + prevBrandHP);
                this._setBrowserHomePage(setupData.HomePage.url, false, setupData.HomePage.force, prevBrandHP);
                
                let prevDefaultQSName;
                try {
                    let osDescription = (prevBrowserConf.QuickSearch && prevBrowserConf.QuickSearch.OpenSearchDescription);
                    if (osDescription)
                        prevDefaultQSName = branding.expandBrandTemplates(osDescription.ShortName, prevBrandMap);
                }
                catch (e) {
                    this._logger.error("Could not get previous QS name. " + e);
                }
                this._writeQuickSearches(null, prevDefaultQSName);
                
                let prevKeywordURL;
                try {
                    if (prevBrowserConf && prevBrowserConf.LocationBarSearch)
                        prevKeywordURL = branding.expandBrandTemplates(prevBrowserConf.LocationBarSearch.Url, prevBrandMap);
                }
                catch (e) {
                    this._logger.error("Could not get previous location bar search URL. " + e);
                }
                this._setLocationBarSearch(setupData.LocationBarSearch.urlForce, prevKeywordURL);
            },
            
            
            _setBrowserHomePage: function Installer__setBrowserHomePage(aHomePageURL, aForce, aBrandForce, aPrevBrandHP) {
                if (!aHomePageURL)
                    return;
                
                let currentHP = this.getBrowserHomePage();
                if (!(aForce || aBrandForce) && (!aPrevBrandHP || aPrevBrandHP !== currentHP))
                    return;
                let setCurrentHP = !aForce && !this.isOverridableURL(currentHP);
                
                let homePageURL = setCurrentHP ? currentHP : aHomePageURL;
                
                Preferences.set("browser.startup.homepage", homePageURL);
                
                if (Preferences.get("browser.startup.page", 1) === 0)
                    Preferences.set("browser.startup.page", 1);
                
                Preferences.set(barApp.name + ".defence.homepage.protected", homePageURL);
                
                this._logger.debug("Changed browser HP from '" + currentHP + "' to '" + homePageURL + "'");
            },
            
            _setLocationBarSearch: function Installer__setLocationBarSearch(aForce, prevKeywordURL) {
                let url = this.setupData.LocationBarSearch.url;
                if (!url)
                    return;
                
                const browserPrefName = "keyword.URL";
                let currentBrowserURL;
                try {
                    currentBrowserURL = Cc["@mozilla.org/preferences-service;1"]
                                            .getService(Ci.nsIPrefService)
                                            .QueryInterface(Ci.nsIPrefBranch2)
                                            .getComplexValue(browserPrefName, Ci.nsIPrefLocalizedString)
                                            .data
                                        || Preferences.get(browserPrefName, null);
                } catch (e) {}
                
                if (!currentBrowserURL) {
                    try {
                        const SBS = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService);
                        let configBundle = SBS.createBundle("chrome://browser-region/locale/region.properties");
                        currentBrowserURL = configBundle.GetStringFromName(browserPrefName);
                    } catch (e) {}
                }
                
                if (!currentBrowserURL) {
                    try {
                        let submissionURL = Cc["@mozilla.org/browser/search-service;1"]
                                                .getService(Ci.nsIBrowserSearchService)
                                                .originalDefaultEngine
                                                .getSubmission("TestQueryString", "application/x-moz-keywordsearch")
                                                .uri.spec;
                        
                        if (submissionURL && /=TestQueryString$/.test(submissionURL))
                            currentBrowserURL = submissionURL.replace(/TestQueryString$/, "");
                        
                    } catch (e) {}
                }
                if (!aForce && (!prevKeywordURL || prevKeywordURL !== currentBrowserURL))
                    return;
                if (!this.isOverridableURL(currentBrowserURL))
                    url = currentBrowserURL;
                
                Preferences.set(browserPrefName, url);
                
                this._logger.debug("Changed keywordURL from '" + currentBrowserURL + "' to '" + url + "'");
            },
            _getDataForDistribution: function Installer__getDataForDistribution() {
                let homePageURL = this.setupData.HomePage.url;
                let currentHomePageURL = this.getBrowserHomePage();
                let homePageEqualDefault = (currentHomePageURL == homePageURL) ||
                                           (!this.isOverridableURL(currentHomePageURL) && !this.isOverridableURL(homePageURL));
                
                let quickSearchers = this._getQuickSearchers();
                let firstQSShortName = quickSearchers[0] && quickSearchers[0].shortName;
                let defaultEngineName = Preferences.get("browser.search.defaultenginename", null);
                let quickSearchEqualDefault =
                        !!(defaultEngineName && (
                            (defaultEngineName == firstQSShortName) ||
                            (!this.isQSOverridable(defaultEngineName) && !this.isQSOverridable(firstQSShortName))
                        ));
                
                return {
                    isHomepageChecked: false,
                    isDefaultHomepage: homePageEqualDefault,
                    HP: { url: homePageURL },
                    
                    isSearchChecked: false,
                    isDefaultQS: quickSearchEqualDefault,
                    QS: quickSearchers
                };
            },
            _getQuickSearchers: function Installer__getQuickSearchers() {
                let qsList = [];
                
                let quickSearches;
                let configXML;
                
                try {
                    configXML = branding.brandPackage.getXMLDocument("/browser/browserconf.xml");
                    quickSearches = configXML.querySelectorAll("Browser > QuickSearch > OpenSearchDescription");
                } catch (e) {
                    Cu.reportError(e);
                    return qsList;
                }

                if (!quickSearches || !quickSearches.length)
                    return qsList;
                
                const DataURI = Cu.import("resource://" + barApp.name + "-mod/DataURI.jsm", {}).DataURI;
                const YB_NS = "http://bar.yandex.ru/";

                for (let i = 0, len = quickSearches.length; i < len; i++) {
                    let qs = quickSearches[i];
                    
                    let uniqName = qs.getAttributeNS(YB_NS, "uniqName");
                    if (!uniqName)
                        continue;
                    
                    let shortName = qs.querySelector("ShortName");
                    shortName = (shortName && shortName.textContent) || "";
                    if (!shortName)
                        continue;
                    
                    let osUrls = qs.querySelectorAll("Url");
                    
                    for (let j = 0, len = osUrls.length; j < len; j++) {
                        let url = osUrls[j];
                        
                        let templateAttr = url.getAttribute("template");
                        if (templateAttr)
                            url.setAttribute("template", branding.expandBrandTemplatesEscape(templateAttr));
                    }
                    
                    let searchURL = "";
                    let searchURLElement = qs.querySelector("Url:not([rel])[type='text/html']");
                    if (searchURLElement && searchURLElement.hasAttribute("template"))
                        searchURL = searchURLElement.getAttribute("template");
                    
                    let suggestURL = "";
                    let suggestURLElement = qs.querySelector("Url[rel='suggestions'][type='application/json']");
                    if (suggestURLElement && suggestURLElement.hasAttribute("template"))
                        suggestURL = suggestURLElement.getAttribute("template");
                    
                    let inputEncoding = qs.querySelector("InputEncoding");
                    inputEncoding = (inputEncoding && inputEncoding.textContent) || "UTF-8";
                    
                    let imageURL;
                    let image = qs.querySelector("Image");
                    if (image) {
                        let imageFile;
                        let imagePath = image.textContent;
                        imageURL = imagePath;
                        if (imagePath && (imageFile = branding.brandPackage.findFile(imagePath))) {
                            image.textContent = DataURI.fromFile(imageFile);
                            imageURL = image.textContent;
                        }
                    }
                    
                    let qsObject = {
                        uniqName: uniqName,
                        isDefault: i == 0,
                        shortName: shortName,
                        image: imageURL,
                        searchURL: searchURL,
                        suggestURL: suggestURL,
                        inputEncoding: inputEncoding
                    };
                    
                    qsList.push(qsObject);
                }
                
                return qsList;
            },
            _writeQuickSearches: function Installer__writeQuickSearches(forceSetDefault, prevDefaultQSName) {
                let searchPluginsDir = Cc["@mozilla.org/file/directory_service;1"]
                                           .getService(Ci.nsIProperties)
                                           .get("ProfD", Ci.nsIFile);
                searchPluginsDir.append("searchplugins");
                
                const QS_FILENAME_PREFIX = "ybqs-";
                
                if (searchPluginsDir.exists() && searchPluginsDir.isDirectory()) {
                    let searchPluginsDirEntries = searchPluginsDir.directoryEntries;
                    while (searchPluginsDirEntries.hasMoreElements()) {
                        let qsFile = searchPluginsDirEntries.getNext().QueryInterface(Ci.nsIFile);
                        if (qsFile.isFile() && qsFile.leafName.indexOf(QS_FILENAME_PREFIX) == 0)
                            fileutils.removeFileSafe(qsFile);
                    }
                }
                
                let quickSearches;
                let configXML;
                
                try {
                    configXML = branding.brandPackage.getXMLDocument("/browser/browserconf.xml");
                    quickSearches = configXML.querySelectorAll("Browser > QuickSearch > OpenSearchDescription");
                } catch (e) {
                    Cu.reportError(e);
                    return;
                }

                if (!quickSearches || !quickSearches.length)
                    return;

                fileutils.forceDirectories(searchPluginsDir);

                const DataURI = Cu.import("resource://" + barApp.name + "-mod/DataURI.jsm", {}).DataURI;
                const YB_NS = "http://bar.yandex.ru/";

                for (let i = 0, len = quickSearches.length; i < len; i++) {
                    let qs = quickSearches[i];
                    
                    let uniqName = qs.getAttributeNS(YB_NS, "uniqName");
                    if (!uniqName)
                        continue;
                    if (i == 0 && forceSetDefault !== false && this.isCurrentQSOverridable()) {
                        let shortName = qs.querySelector("ShortName");
                        shortName = (shortName && shortName.textContent) || "";
                        
                        let browserDefaultEngineName = Preferences.get("browser.search.defaultenginename", "");
                        if (forceSetDefault ||
                            (shortName != prevDefaultQSName && prevDefaultQSName === browserDefaultEngineName)) {
                            if (forceSetDefault)
                                Preferences.set("browser.search.selectedEngine", shortName);
                            else
                                brandPrefs.set("qsEngineNameForSelect", shortName);
                            
                            this._logger.debug("Changed browser QS to '" + shortName + "'");
                            
                            Preferences.set("browser.search.defaultenginename", shortName);
                        }
                    }
                    
                    let qsFile = searchPluginsDir.clone();
                    qsFile.append(QS_FILENAME_PREFIX + uniqName + ".xml");
                    
                    if (uniqName == "yandex" && this._copyYandexQSFromApplication(qsFile))
                        continue;
                    
                    let osUrls = qs.querySelectorAll("Url");
                    
                    for (let j = 0, len = osUrls.length; j < len; j++) {
                        let url = osUrls[j];
                        
                        let templateAttr = url.getAttribute("template");
                        if (templateAttr)
                            url.setAttribute("template", branding.expandBrandTemplatesEscape(templateAttr));
                        
                        let typeAttr = url.getAttribute("type");
                        if (typeAttr == "application/json")
                            url.setAttribute("type", "application/x-suggestions+json");
                    }
                    let searchFormURLElement = qs.querySelector("Url[rel='search-form'][type='text/html']");
                    if (searchFormURLElement) {
                        let searchFormURL = searchFormURLElement.getAttribute("template");
                        searchFormURLElement.parentNode.removeChild(searchFormURLElement);
                        
                        let searchFormElement = configXML.createElement("SearchForm");
                        searchFormElement.textContent = searchFormURL;
                        qs.appendChild(searchFormElement);
                    }
                    
                    let image = qs.querySelector("Image");
                    if (image) {
                        let imageFile;
                        let imagePath = image.textContent;
                        if (imagePath && (imageFile = branding.brandPackage.findFile(imagePath)))
                            image.textContent = DataURI.fromFile(imageFile);
                    }
                    
                    fileutils.writeTextFile(qsFile, xmlutils.serializeXML(qs));
                }
            },
            _copyYandexQSFromApplication: function Installer__copyYandexQSFromApplication(aFile) {
                let curProcDir;

                try {
                  curProcDir = Cc["@mozilla.org/file/directory_service;1"]
                                   .getService(Ci.nsIProperties)
                                   .get("CurProcD", Ci.nsIFile);
                } catch(e) {
                    return false;
                }
                
                let yandexQSFile = curProcDir.clone();
                ["distribution", "searchplugins", "common", "yasearch.xml"]
                .forEach(function(aPath) {
                    yandexQSFile.append(aPath);
                });
                
                if (!yandexQSFile.exists() || !yandexQSFile.isFile()) {
                    yandexQSFile = curProcDir.clone();
                    
                    ["searchplugins", "yandex.xml"]
                    .forEach(function(aPath) {
                        yandexQSFile.append(aPath);
                    });
                }
                
                if (yandexQSFile.exists() && yandexQSFile.isFile()) {
                    try {
                        yandexQSFile.copyTo(aFile.parent, aFile.leafName);
                        return true;
                    } catch(e) {}
                }
                
                return false;
            },
            
            get _licenseDirURL() {
                return "chrome://" + barApp.name + "/content/custombar/dialogs/license/";
            },
            _showLicenseWindow: function Installer__showLicenseWindow() {
                try {
                    let windowWatcher = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
                    
                    let args = this.setupData;
                    args.wrappedJSObject = args;
                    
                    let setupWin = windowWatcher.openWindow(null, this._licenseDirURL + "wizard.xul",
                                                            null, "centerscreen,modal,resizable", args);
                    
                    return args.License.checked;
                } catch (e) {
                    Cu.reportError(e);
                }
                
                return true;
            },
            _disableAddonOnLicenseRefuse: function Installer__disableAddonOnLicenseRefuse() {
                let windowWatcher = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
                let refuseWin = windowWatcher.openWindow(null, this._licenseDirURL + "refused.xul",
                                                         null, "centerscreen,modal", null);
                barApp.addonManager.disableAddon();
            },
            
            get _logger() {
                delete this._logger;
                return this._logger = barApp.getLogger("Branding.Installer");
            }
        };
        
        delete this.Installer;
        return this.Installer = Installer;
    }
};

const packageDownloader = {
    init: function packageDownloader_init(barApp) {
        this._barApp = barApp;
        this._logger = barApp.getLogger("Branding");
        
        let updTimePref = barApp.preferences.get(this._consts.CHECK_TS_PREF_NAME, 0);
        this._lastCheckTimestamp = parseInt(updTimePref, 10) || 0;
    },
    
    sheduleUpdates: function packageDownloader_sheduleUpdates(URL) {
        this._URL = URL;
        
        if (!this._timer) {
            let nextCheckInterval = this._consts.UPD_INTERVAL * 1000 - Math.abs(this._lastCheckTimestamp * 1000 - Date.now());
            nextCheckInterval = Math.max(nextCheckInterval, this._barApp.generateDelay() * 1000);
            
            this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
            this._timer.initWithCallback(this, nextCheckInterval, this._timer.TYPE_ONE_SHOT);
        }
    },
    
    stopUpdates: function packageDownloader_stopUpdates() {
        if (this._timer)
            this._timer.cancel();
        
        if (this._currWBPersist)
            this._currWBPersist.cancelSave();
    },
    notify: function packageDownloader_notify() {
        try {
            this._checkUpdate();
        }
        catch (e) {
            this._logger.error("Could not check partner package update. \n" + strutils.formatError(e));
        }
        this._timer.initWithCallback(this, this._consts.UPD_INTERVAL * 1000, this._timer.TYPE_ONE_SHOT);
    },
    
    _consts: {
        CHECK_TS_PREF_NAME: "branding.lastcheck",
        UPD_INTERVAL: 3 * 24 * 60 * 60
    },
    _barApp: null,
    _timer: null,
    _URL: undefined,
    _lastCheckTimestamp: 0,
    _currWBPersist: null,
    _currTempFile: null,
    
    _checkUpdate: function packageDownloader__checkUpdate() {
        this._logger.debug("Trying to update from " + this._URL);
        
        this._lastCheckTimestamp = Math.round(Date.now() / 1000);
        this._barApp.preferences.set(this._consts.CHECK_TS_PREF_NAME, this._lastCheckTimestamp);
        
        let packageURI = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService).newURI(this._URL, null, null);
        let wbPersist = Cc["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Ci.nsIWebBrowserPersist);
        
        this._currWBPersist = wbPersist;
        
        wbPersist.persistFlags = wbPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES |
                                 wbPersist.PERSIST_FLAGS_IGNORE_REDIRECTED_DATA |
                                 wbPersist.PERSIST_FLAGS_CLEANUP_ON_FAILURE |
                                 wbPersist.PERSIST_FLAGS_BYPASS_CACHE;
        wbPersist.progressListener = this;
        
        let httpHeaders = "If-Modified-Since: " + branding.brandingDate.toUTCString() + "\r\n";
        this._logger.debug("HEADERS: " + httpHeaders);
        
        this._currTempFile = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("TmpD", Ci.nsIFile);
        this._currTempFile.append(misc.CryptoHash.getFromString(this._URL, "MD5"));
        this._currTempFile.createUnique(Ci.nsIFile.NORMAL_FILE_TYPE, parseInt("0755", 8));
        
        wbPersist.saveURI(packageURI, null, null, null, httpHeaders, this._currTempFile);
    },
    
    onLocationChange: function packageDownloader_onLocationChange() {},
    onSecurityChange: function packageDownloader_onSecurityChange() {},
    onStatusChange: function packageDownloader_onStatusChange() {},
    
    onStateChange: function packageDownloader_onStateChange(webProgress, request, state, status) {
        try {
            if (state & Ci.nsIWebProgressListener.STATE_STOP) {
                this._currWBPersist = null;
                
                let httpChannel = request.QueryInterface(Ci.nsIHttpChannel);
                let httpStatus = httpChannel.responseStatus;
                
                this._logger.debug("Branding package request finished. Status: " + status + ". HTTP: " + httpStatus);
                
                if ((status == Cr.NS_OK) && (httpStatus == 200)) {
                    let dateHeader;
                    let responseDate;
                    try {
                        dateHeader = httpChannel.getResponseHeader("Date");
                        responseDate = new Date(dateHeader);
                    } catch(e) {
                        this._logger.warn("Could not get server Date header. \n" + strutils.formatError(e));
                    }
                    
                    if (!responseDate || responseDate == "Invalid Date") {
                        this._logger.warn("Could not parse server Date header: " + dateHeader);
                        responseDate = new Date();
                    }
                    
                    branding.newPackageReady(this._currTempFile, responseDate);
                }
            }
        }
        catch (e) {
            this._logger.warn("Web progress state change handler failed. \n" + strutils.formatError(e));
        }
    },
    
    onProgressChange: function packageDownloader_onProgressChange(webProgress, request, currentSelfProgress,
                                                                  maxSelfProgress, currentTotalProgress,
                                                                  maxTotalProgress) {
    }
};
const xbDataProvider = {
    init: function xbBrandProv_init(application) {
        this._application = application;
        this._protocolHandler = application.core.xbProtocol;
        this._protocolHandler.setDataProvider(this._domain, this);
    },
    
    finalize: function xbBrandProv_finalize() {
        this._application = null;
        this._protocolHandler.setDataProvider(this._domain, null);
    },
    
    packageUpdated: function xbBrandProv_packageUpdated(productInfo) {
        this._namesStr = undefined;
        this._pathStr = undefined;
    },
    
    get wrappedJSObject() {
        return this;
    },
    get UUID() {
        return this._domain;
    },
    
    newChannel: function xbBrandProv_newChannel(aURI) {
        let dataString = this._getData(aURI);
        var inputStream = strutils.utf8Converter.convertToInputStream(dataString);
        
        let channel = Cc["@mozilla.org/network/input-stream-channel;1"]
                          .createInstance(Ci.nsIInputStreamChannel)
                          .QueryInterface(Ci.nsIChannel);
        channel.setURI(aURI);
        channel.originalURI = aURI;
        channel.contentStream = inputStream;
        
        return channel;
    },
    _protocolHandler: null,
    _domain: "branding",
    _pathStr: undefined,
    _namesStr: undefined,
    _namesTpl: "\
        <!ENTITY product1.nom \"{product1.nom}\">\
        <!ENTITY product1.gen \"{product1.gen}\">\
        <!ENTITY product1.dat \"{product1.dat}\">\
        <!ENTITY product1.acc \"{product1.acc}\">\
        <!ENTITY product1.ins \"{product1.ins}\">\
        <!ENTITY product1.pre \"{product1.pre}\">\
        <!ENTITY product1.loc \"{product1.loc}\">\
        \
        <!ENTITY product2.nom \"{product2.nom}\">\
        <!ENTITY product2.gen \"{product2.gen}\">\
        <!ENTITY product2.dat \"{product2.dat}\">\
        <!ENTITY product2.acc \"{product2.acc}\">\
        <!ENTITY product2.ins \"{product2.ins}\">\
        <!ENTITY product2.pre \"{product2.pre}\">\
        <!ENTITY product2.loc \"{product2.loc}\">\
        \
        <!ENTITY vendor.nom \"{vendor.nom}\">\
        <!ENTITY vendor.gen \"{vendor.gen}\">\
        <!ENTITY vendor.dat \"{vendor.dat}\">\
        <!ENTITY vendor.acc \"{vendor.acc}\">\
        <!ENTITY vendor.ins \"{vendor.ins}\">\
        <!ENTITY vendor.pre \"{vendor.pre}\">\
        <!ENTITY vendor.loc \"{vendor.loc}\">",
    
    _getData: function xbBrandProv__getData(aURI) {
        if (aURI.path == "/names.dtd") {
            if (!this._namesStr)
                this._namesStr = branding.expandBrandTemplates(this._namesTpl);
            return this._namesStr;
        }
        
        if (aURI.path == "/path.dtd") {
            if (!this._pathStr)
                this._pathStr = "<!ENTITY xb.branding.path \"" + branding.brandPackage.resolvePath(".") + "\">";
            return this._pathStr;
        }
        
        throw Cr.NS_ERROR_MALFORMED_URI;
    }
};
