"use strict";

const EXPORTED_SYMBOLS = ["daylystat"];

const { classes: Cc, interfaces: Ci, utils: Cu, results: Cr } = Components;
const GLOBAL = this;

const daylystat = {
    init: function DaylyStat_init(barApplication) {
        this._application = barApplication;
        
        barApplication.core.Lib.sysutils.copyProperties(barApplication.core.Lib, GLOBAL, {
            sysutils: 1,
            strutils: 1,
            misc: 1
        });
        
        this._logger = barApplication.getLogger("DaylyStat");
    },
    
    finalize: function DaylyStat_finalize() {
        if (this._timer)
            this._timer.cancel();
    },
    
    onFisrtNavigatorReady: function DaylyStat_onFisrtNavigatorReady() {
        this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
        let chkTimePref = this._application.preferences.get(this._consts.LAST_CHK_PREF_NAME, 0);
        this._lastCheckTime = parseInt(chkTimePref, 10) || 0;
        
        let nextCheckInterval = this._consts.CHK_INTERVAL * 1000 - Math.abs(this._lastCheckTime * 1000 - Date.now());
        nextCheckInterval = Math.max(nextCheckInterval, 0);
        
        this._timer.initWithCallback(this, nextCheckInterval, this._timer.TYPE_ONE_SHOT);
    },
    
    notify: function DaylyStat_notify() {
        this._postStatistics();
        this._rememberTry();
        this._timer.initWithCallback(this, this._consts.CHK_INTERVAL * 1000, this._timer.TYPE_ONE_SHOT);
    },
    
    _consts: {
        LAST_CHK_PREF_NAME: "daylystat.sent",
        CHK_INTERVAL: 24 * 60 * 60
    },
    
    _rememberTry: function DaylyStat__rememberTry() {
        let secondsSinceUnixEpoch = Math.floor(Date.now() / 1000);
        this._application.preferences.set(this._consts.LAST_CHK_PREF_NAME, secondsSinceUnixEpoch);
    },
    
    _postStatistics: function DaylyStat__postStatistics() {
        try {
            this._postActiveComps();
        }
        catch (e) {
            this._logger.error("Could not post current buttons statistics. " + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
        
        try {
            this._postDefPresetDigest();
        }
        catch (e) {
            this._logger.error("Could not post default preset digest. " + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
        
    },
    
    _postActiveComps: function DaylyStat__postActiveComps() {
        let statLogger = this._application.statLogger;
        const DAY_USE_CODE = statLogger.ACTIONS.COMP_DAYUSE;
        
        let currPluginsState = this._application.widgetLibrary.getCurrentPluginsState();
        for each (let [pluginID, activated] in Iterator(currPluginsState)) {
            if (activated)
                statLogger.logSysAction(pluginID, DAY_USE_CODE);
        }
        
        let browserWindow = misc.getTopBrowserWindow();
        if (!browserWindow)
            return;
        let wndController = browserWindow[this._application.name + "OverlayController"];
        let currentWidgetItems = wndController.getAllWidgetItems();
        currentWidgetItems.forEach(function extractID(widgetElement) {
            if (widgetElement.wdgtIsXBWidget) {
                statLogger.logSysAction(widgetElement.wdgtPrototypeID, DAY_USE_CODE);
                return;
            }
            let staticWidgetID = this._findSWIDByOldID(widgetElement.id);
            if (staticWidgetID)
                statLogger.logSysAction(staticWidgetID, DAY_USE_CODE);
        }, this);
    },
    
    _findSWIDByOldID: function DaylyStat__findSWIDByOldID(oldID) {
        for each (let [commonID, itemName] in Iterator(this._staticWidgetsInfo))
            if (oldID == itemName)
                return commonID;
        return undefined;
    },
    
    get _staticWidgetsInfo() {
        let staticWidgetsInfo = this._application.overlayProvider.staticWidgets;
        this.__defineGetter__("_staticWidgetsInfo", function getSWI() staticWidgetsInfo);
        return this._staticWidgetsInfo;
    },
    
    _postDefPresetDigest: function DaylyStat__postDefPresetDigest() {
        let allPresetIDs = [];
        for each (let presetEntry in this._application.defaultPreset.allEntries)
            allPresetIDs.push(presetEntry.componentID);
        let allIDsSpaceSeparated = allPresetIDs.join(" ");
        let statLogger = this._application.statLogger;
        statLogger.logSysAction(allIDsSpaceSeparated, statLogger.ACTIONS.DP_HASH);
    }
};
