"use strict";
function NativeBarAPI(componentInfo  , logger) {
    if (!componentInfo
            || !(componentInfo.package_ instanceof BarPlatform.ComponentPackage))
        throw new CustomErrors.EArgType("componentInfo", "XBComponentInfo", componentInfo);
    this._componentInfo = componentInfo;
    
    if (!logger)
        throw new CustomErrors.EArgType("logger", "Object", logger);
    this._logger = logger;
    [ "Environment", "Controls", "Statistics", "Settings", "Package",
      "Files", "XMLUtils", "StrUtils", "Browser", "Network",
      "Services", "Autocomplete", "DistrData"
    ].forEach(function inst(apiName) {
        sysutils.defineLazyGetter(this, apiName, function lazyGetter() {
            return new NativeBarAPI[apiName](this._componentInfo, this._logger);
        });
    }, this);
};

function NoCompInfo(package_) {
    this._package = package_;
}

NoCompInfo.prototype = {
    get id() {
        throw new Error(this._NO_COMP_INFO);
    },
    
    get type() {
        throw new Error(this._NO_COMP_INFO);
    },
    
    get component() {
        throw new Error(this._NO_COMP_INFO);
    },
    
    get package_() {
        return this._package;
    },
    
    _NO_COMP_INFO: "Shareable API can't access component information!"
};

NativeBarAPI.prototype = {
    get componentID() {
        return this._componentInfo.id;
    },
    
    get logger() {
        return this._logger;
    },
    
    
    queryAPIVersion: function NativeAPI_queryAPIVersion(version) {
        let apiVer = Number(version);
        if ((apiVer > 0) && (apiVer <= this.Environment.barPlatform.version))
            return this;
        throw new Error(strutils.formatString("API version %1 is not available", [version]));
    },
    
    
    get shareableAPI() {
        let logger = Log4Moz.repository.getLogger(appCore.appName + ".NativeAPI." + this._componentInfo.package_.id);
        let shareableAPI = new NativeBarAPI(new NoCompInfo(this._componentInfo.package_), logger);
        this.__defineGetter__("shareableAPI", function() shareableAPI);
        return this.shareableAPI;
    },
    finalize: function NativeAPI_finalize() {
        try {
            NativeComponents._unregisterServices(this.componentID);
        }
        catch (e) {
            this._logger.error("NativeComponents._unregisterServices failed. " + strutils.formatError(e));
        }
        try {
            NativeComponents._releaseServices(this.componentID);
        }
        catch (e) {
            this._logger.error("NativeComponents._releaseServices failed. " + strutils.formatError(e));
        }
        
        this.Settings._finalize();
        this.Autocomplete._finalize();
        this.Browser._finalize();
        
        this._componentInfo = null;
    }
};

NativeBarAPI.Services = function Services(componentInfo) {
    this._componentInfo = componentInfo;
};

NativeBarAPI.Services.prototype = {
    registerService: function NativeAPI_registerService(serviceName, serviceObject) {
        return NativeComponents.registerService(this._componentInfo.id, serviceName, serviceObject);
    },
    
    notifyServiceUsers: function NativeAPI_notifyServiceUsers(serviceName, topic, data) {
        NativeComponents.notifyServiceUsers(this._componentInfo.id, serviceName, topic, data);
    },
    
    unregisterService: function NativeAPI_unregisterService(serviceName) {
        return NativeComponents.unregisterService(this._componentInfo.id, serviceName);
    },
    obtainService: function NativeAPI_obtainService(providerID, serviceName, eventListener) {
        return NativeComponents.obtainService(providerID, serviceName, eventListener, this._componentInfo.id);
    },
    
    releaseService: function NativeAPI_releaseService(providerID, serviceName, eventListener) {
        NativeComponents.releaseService(providerID, serviceName, eventListener);
    }
};

NativeBarAPI.Environment = function Environment() {
};
NativeBarAPI.Environment.prototype = {
    constructor: NativeBarAPI.Environment,
    
    os: {
        get name() {
            return sysutils.platformInfo.os.name;
        }
        
    },
    
    browser: {
        get name() {
            return sysutils.platformInfo.browser.name;
        },
        
        
        get version() {
            return sysutils.platformInfo.browser.version;
        }
    },
    
    addon: {
        get id() {
            let id = AddonManager.getAddonId(appCore.extensionPathFile);
            this.__defineGetter__("id", function NativeAPI_Env_addon_id() id);
            return this.id;
        },
        
        get version() {
            let version = AddonManager.getAddonVersion(appCore.extensionPathFile);
            this.__defineGetter__("version", function NativeAPI_Env_addon_version() version);
            return this.version;
        },
        
        
        get locale() {
            let localeString = application.localeString;
            let localeComponents = misc.parseLocale(localeString);
            
            let locale = {
                toString: function NativeAPI_locale_toString() localeString,
                language: localeComponents.language || "",
                country:  localeComponents.country || "",
                region:   localeComponents.region || ""
            };
            
            this.__defineGetter__("locale", function NativeAPI_locale() locale);
            return this.locale;
        },
        
        
        get userID() {
            let oldCore = application.core.oldCore;
            return (oldCore && oldCore.guidString) || "";
        },
        
        
        get goingToUninstall() {
            return application.addonManager.isAddonUninstalling;
        }
    },
    
    barPlatform: {
        get name() {
            return appCore.appName;
        },
        
        get version() {
            return appCore.version;
        }
    }
};

NativeBarAPI.Environment.__proto__ = NativeBarAPI.Environment.prototype;
NativeBarAPI.Controls = function Controls(componentInfo, logger) {
    this._componentInfo = componentInfo;
    this._logger = logger;
};

NativeBarAPI.Controls.prototype = {
    
    openSettingsDialog: function NativeAPI_openSettingsDialog(browserWindow, setupID, paneType) {
        application.openSettingsDialog(browserWindow, setupID, paneType);
    },
    
    
    openAboutDialog: function NativeApi_openAboutDialog() {
        application.openAboutDialog();
    },
    
    
    logUsageStat: function NativeAPI_logUsageStat(eventID) {
        this._logger.warn("NativeAPI.Controls.logUsageStat is deprecated!");
        application.statLogger.logShortAction(eventID);
    },
    
    
    navigateBrowser: function NativeAPI_navigateBrowser(navigateData) {
        return application.navigate(navigateData);
    },
    
    
    addWidget: function NativeAPI_addWidget(widgetID, browserWindow, relativeTo, placeAfter) {
        if (!browserWindow)
            browserWindow = misc.getTopBrowserWindow();
        else if (!(browserWindow instanceof Ci.nsIDOMWindow))
            throw new TypeError("XUL window required " + browserWindow);
        
        let wndController = NativeComponents._getWindowController(browserWindow);
        widgetID = widgetID || this._componentInfo.id;
        let [, widgetElement] = wndController.placeWidget(widgetID, relativeTo, placeAfter, undefined, true);
        return [widgetElement.wdgtInstanceID, widgetElement];
    },
    
    removeWidget: function NativeAPI_removeWidget(WIIDorElement, browserWindow) {
        if (!browserWindow) {
            if (WIIDorElement instanceof Ci.nsIDOMElement)
                browserWindow = WIIDorElement.ownerDocument.defaultView;
            else
                browserWindow = misc.getTopBrowserWindow();
        }
        if (!(browserWindow instanceof Ci.nsIDOMWindow))
            throw new TypeError("XUL window required " + browserWindow);
        NativeComponents._getWindowController(browserWindow).removeItem(WIIDorElement);
    },
    
    enablePlugin: function NativeAPI_enablePlugin(pluginID) {
        application.widgetLibrary.getPlugin(pluginID).enable();
    },
    
    disablePlugin: function NativeAPI_disablePlugin(pluginID) {
        application.widgetLibrary.getPlugin(pluginID).disable();
    }
};


NativeBarAPI.Statistics = function Statistics(componentInfo, logger) {
    this._componentInfo = componentInfo;
    this._logger = logger;
    this._statLogger = application.statLogger;
};

NativeBarAPI.Statistics.prototype = {
    
    logShortAction: function NativeAPI_logShortAction(eventID) {
        this._statLogger.logShortAction(eventID);
    },
    
    logCustomAction: function NativeAPI_logCustomAction(eventID) {
        this._statLogger.logCustomAction(this._componentInfo.id, eventID);
    },
    
    logButtonClick: function NativeAPI_logButtonClick() {
        this._statLogger.logSysAction(this._componentInfo.id, this._statLogger.ACTIONS.BTN_CLICK);
    },
    
    logMenuClick: function NativeAPI_logMenuClick() {
        this._statLogger.logSysAction(this._componentInfo.id, this._statLogger.ACTIONS.MENU_CLICK);
    },
    
    
    logAddonEvents: function NativeAPI_logAddonEvents(eventsMap) {
        this._statLogger.logAddonEvents(eventsMap);
    },
    
    
    fetchBarNavigStat: function NativeAPI_fetchBarNavigStat() {
        return this._statLogger.readActions();
    },
    
    
    get barnavigR1String() {
        let oldCore = application.core.oldCore;
        return (oldCore && oldCore.barnavigR1String) || "";
    },
    
    
    set barnavigR1String(aR1String) {
        if (typeof aR1String != "string")
            throw new TypeError("aR1String must be a string");
        
        let oldCore = application.core.oldCore;
        if (oldCore)
            oldCore.barnavigR1String = aR1String;
    }
};

NativeBarAPI.Settings = function Settings(componentInfo, logger) {
    this._componentInfo = componentInfo;
    this._logger = logger;
    this._settingsObservers = {};
};

NativeBarAPI.Settings.prototype = {
    getValue: function NativeAPI_getValue(settingName, WIID) {
        if (this._componentInfo.type == "widget")
            return this._componentInfo.component.lookupGetSettingValue(settingName, WIID);
        return this._componentInfo.component.getSettingValue(settingName);
    },
    
    
    setValue: function NativeAPI_setValue(settingName, newValue, WIID) {
        if (this._componentInfo.type == "widget")
            this._componentInfo.component.lookupSetSettingValue(settingName, WIID, newValue);
        else
            this._componentInfo.component.applySetting(settingName, newValue);
    },
    
    observeChanges: function NativeAPI_observeChanges(observer, WIID) {
        if (!observer || (typeof observer.onSettingChange != "function"))
            throw new TypeError("Observer must be an object with 'onSettingChange' method");
        
        let ci = this._componentInfo;
        if (ci.type == "plugin")
            WIID = ci.component.id;
        else if (WIID)
            this._watchInstSettings(WIID);
        this._watchSettings(WIID);
        
        this._settingsObservers[WIID] = observer;
    },
    
    ignoreChanges: function NativeAPI_ignoreChanges(WIID) {
        let ci = this._componentInfo;
        if (ci.type == "plugin")
            WIID = ci.component.id;
        else
            this._ignoreInstSettings(WIID);
        
        delete this._settingsObservers[WIID];
        if (sysutils.isEmptyObject(this._settingsObservers))
            this._ignoreSettings();
    },
    
    get PrefsModule() {
        return Preferences;
    },
    
    
    getPackageBranchPath: function NativeAPI_getPackageBranchPath() {
        return this._pbp || (this._pbp = NativeComponents.makePackagePrefPath(this._componentInfo.package_.id));
    }, 
    
    
    getComponentBranchPath: function NativeAPI_getComponentBranchPath() {
        return this._cbp || (this._cbp = NativeComponents.makeWidgetPrefPath(this._componentInfo.id));
    },
    
    
    getInstanceBranchPath: function NativeAPI_getInstanceBranchPath(WIID) {
        return NativeComponents.makeInstancePrefPath(this._componentInfo.id, WIID);
    },
    
    
    getStaticBranchPath: function NativeAPI_getStaticBranchPath(WIID) {
        return NativeComponents.makeStaticBranchPath(this._componentInfo.id, WIID);
    },
    
    
    get addonBranchPath() {
        let path = "extensions." + NativeBarAPI.Environment.addon.id + ".";
        this.__defineGetter__("addonBranchPath", function NativeAPI_Settings_addonBranchPath() path);
        return this.addonBranchPath;
    },
    
    _settingsObservers: null,
    _watchingCommonSettings: false,
    
    _finalize: function () {
        for (let WIID in this._settingsObservers) {
            try {
                this._ignoreInstSettings(WIID);
            }
            catch (e) {}
        }
        this._ignoreSettings();
        this._settingsObservers = null;
    },
    
    observe: function NativeAPI_observe(subject, topic, data) {
        try {
            switch (topic) {
                case "nsPref:changed": {
                    this._notifySettingsObservers(subject, data);
                    break;
                }
                default:
                    return;
            }
        }
        catch (e) {
            this._logger.error("nsIObserver.observe failed. " + strutils.formatError(e));
        }
    },
    
    _watchSettings: function NativeAPI__watchSettings() {
        if (this._watchingCommonSettings)
            return;
        let component = this._componentInfo.component;
        Preferences.observe2(NativeComponents.makePackagePrefPath(component.pkg.id), this);
        Preferences.observe2(NativeComponents.makeWidgetPrefPath(component.id), this);
    },
    
    _watchInstSettings: function NativeAPI__watchInstSettings(WIID) {
        if (WIID in this._settingsObservers)
            throw new Error("Widget instance " + WIID + " settings are already being watched");
        Preferences.observe2(NativeComponents.makeInstancePrefPath(this._componentInfo.component.id, WIID), this);
    },
    
    _ignoreSettings: function NativeAPI__ignoreSettings() {
        let component = this._componentInfo.component;
        Preferences.ignore2(NativeComponents.makePackagePrefPath(component.pkg.id), this);
        Preferences.ignore2(NativeComponents.makeWidgetPrefPath(component.id), this);
        this._watchingCommonSettings = false;
    },
    
    _ignoreInstSettings: function NativeAPI__ignoreInstSettings(WIID) {
        Preferences.ignore2(NativeComponents.makeInstancePrefPath(this._componentInfo.component.id, WIID), this);
    },
    
    _notifySettingsObservers: function NativeAPI__notifySettingsObservers(prefBranch, prefPath) {
        let component = this._componentInfo.component;
        let prefProperties = BarPlatform.parsePrefPath(prefPath, component.id);
        let settingName = prefProperties.settingName;
        
        if (this._componentInfo.type == "widget") {
            let widgetHasInstSetting = (settingName in component._instSettings );
            if (prefProperties.isInstancePref && !widgetHasInstSetting)
                throw new Error(strutils.formatString("Widget %1 does not have setting named \"%2\"", [component.id, settingName]));
            if ( (!prefProperties.isInstancePref && widgetHasInstSetting) ||
                 (prefProperties.isPackagePref && (settingName in component.widgetSettings)) )
                    return;
        }
        else {
            if (prefProperties.isInstancePref)
                throw new Error("Got instance setting change, but this component is not a widget: " + prefPath);
            if (prefProperties.isPackagePref && (settingName in component.pluginSettings))
                return;
        }
        let newValue = this.getValue(settingName, prefProperties.instanceID);
        for (let [id, observer] in Iterator(this._settingsObservers)) {
            try {
                observer.onSettingChange(settingName, newValue, prefProperties.instanceID);
            }
            catch (e) {
                this._logger.error("Could not notify native core about setting change. " + strutils.formatError(e));
            }
        };
    }
};

NativeBarAPI.Package = function Package(componentInfo, logger) {
    this._package = componentInfo.package_;
    this._logger = logger;
};

NativeBarAPI.Package.prototype = {
    constructor: NativeBarAPI.Package,
    get id() {
        return this._package.id;
    },
    
    
    resolvePath: function NativeAPI_resolvePath(path, basePath) {
        return this._package.resolvePath(path, basePath);
    },
    getFileInputChannel: function NativeAPI_getFileInputChannel(path) {
        return this._package.newChannelFromPath(path);
    }
};

NativeBarAPI.Files = function Files(componentInfo, logger) {
    this._componentInfo = componentInfo;
    this._logger = logger;
};

NativeBarAPI.Files.prototype = {
    getWidgetStorage: function NativeAPI_getWidgetStorage(create ) {
        let result = NativeComponents.getComponentStorage(this._componentInfo.id);
        
        this.getWidgetStorage = function NativeAPI_getWidgetStorage(inCreate) {
            if (inCreate)
                fileutils.forceDirectories(result);
            return result.clone();
        };
        
        return this.getWidgetStorage(create);
    },
    
    getInstanceStorage: function NativeAPI_getInstanceStorage(WIID, create ) {
        let result = this.getWidgetStorage();
        result.append(WIID);
        if (create)
            fileutils.forceDirectories(result);
        return result;
    },
    
    forceDirectories: function NativeAPI_forceDirectories(file) {
        fileutils.forceDirectories(file);
    },
    
    writeStreamToFile: function NativeAPI_writeStreamToFile(inputStream, destFile, accessRights, modeFlags) {
        return fileutils.writeStreamToFile(inputStream, destFile, accessRights, modeFlags);
    },
    
    
    readTextFile: function NativeAPI_readTextFile(file) {
        return fileutils.readTextFile(file);
    },
    
    
    writeTextFile: function NativeAPI_writeTextFile(file, text, accessRights, modeFlags) {
        return fileutils.writeTextFile(file, text, accessRights, modeFlags);
    }
};

NativeBarAPI.XMLUtils = function XMLUtils(componentInfo, logger) {
    this._logger = logger;
};

NativeBarAPI.XMLUtils.prototype = {
    constructor: NativeBarAPI.XMLUtils,
    
    xmlDocFromStream: function NativeAPI_xmlDocFromStream(stream, docURI, baseURI, privileged) {
        return fileutils.xmlDocFromStream(stream, docURI, baseURI, privileged);
    },
    
    xmlDocFromFile: function NativeAPI_xmlDocFromFile(localFile, privileged) {
        return fileutils.xmlDocFromFile(localFile, privileged);
    },

    
    xmlDocToFile: function NativeAPI_xmlDocToFile(xmlDocument, destFile, accessRights, modeFlags) {
        return fileutils.xmlDocToFile(xmlDocument, destFile, accessRights, modeFlags);
    },

    
    transformXMLToFragment: function NativeAPI_transformXMLToFragment(sourceNode, stylesheet, destDoc, oParams) {
        if (!destDoc) {
            let browserWindow = misc.getTopBrowserWindow();
            destDoc = browserWindow.document.implementation.createDocument("", "", null);
        }
        return xmlutils.transformXMLToFragment(sourceNode, stylesheet, destDoc, oParams);
    },
    
    
    queryXMLDoc: function NativeAPI_queryXMLDoc(xpathExpr, contextNode, extNSResolver) {
        return xmlutils.queryXMLDoc(xpathExpr, contextNode, extNSResolver);
    },
    
    getDOMParser: function NativeAPI_getDOMParser(docURI, baseURI, withSystemPrincipal) {
        return xmlutils.getDOMParser(docURI, baseURI, withSystemPrincipal);
    }
};

NativeBarAPI.StrUtils = function StrUtils(componentInfo, logger) {
    this._logger = logger;
};

NativeBarAPI.StrUtils.prototype = {
    constructor: NativeBarAPI.StrUtils,
    
    readStringFromStream: function NativeAPI_readStringFromStream(inputStream) {
        return fileutils.readStringFromStream(inputStream);
    },
    formatError: function NativeAPI_formatError(error) {
        this._logger.error("NativeAPI.StrUtils.formatError is obsolete.");
        return strutils.formatError(error);
    },
    
    
    dumpValue: function NativeBarAPI_dumpValue(value, depth) {
        return sysutils.dump(value, depth);
    }
};

NativeBarAPI.Browser = function Browser(componentInfo, logger) {
    this._componentInfo = componentInfo;
    this._logger = logger;
};

NativeBarAPI.Browser.prototype = {
    getWindowListener: function NativeAPI_getWindowListener(window) {
        return NativeComponents._getWindowController(window).windowListener;
    },
    
    getWindowDataIsland: function NativeAPI_getWindowDataIsland(window) {
        return NativeComponents._getWindowController(window).windowDataIsland;
    },
    
    
    get globalHiddenWindow() {
        return misc.hiddenWindows.getWindow(this.HIDDEN_CHROME_WINDOW_URL);
    },
    
    
    getHiddenFrame: function NativeAPI_getHiddenFrame() {
        this._hiddenFrameAccessed = true;
        return misc.hiddenWindows.getFrame(this._componentInfo.id, this.HIDDEN_CHROME_WINDOW_URL);
    },
    
    
    removeHiddenFrame: function NativeAPI_removeHiddenFrame() {
        return misc.hiddenWindows.removeFrame(this._componentInfo.id);
    },
    
    HIDDEN_CHROME_WINDOW_URL: "chrome://" + application.name + "/content/custombar/overlay/hiddenwindow.xul",
    
    _hiddenFrameAccessed: false,
    
    _finalize: function NativeAPI_Browser__finalize() {
        if (this._hiddenFrameAccessed)
            this.removeHiddenFrame();
    }
};


NativeBarAPI.Network = function Network() {
    
};

NativeBarAPI.Network.prototype = {
    
    Cookie: netutils.Cookie,
    
    
    findCookieValue: function Network_findCookieValue(URLorURI, cookieName, incHttpOnly, checkExpired, strictMatch) {
        return netutils.findCookieValue.apply(netutils, arguments);
    },
    
    
    findCookies: function Network_findCookies(URLorURI, cookieName, incHttpOnly, checkExpired, strictMatch) {
        return netutils.findCookies.apply(netutils, arguments);
    },
    
    
    getCookiesFromHost: function Network_getCookiesFromHost(aHost) {
        return netutils.getCookiesFromHost.apply(netutils, arguments);
    },
    
    
    cookieMatchesURI: function Network_cookieMatchesURL(cookie, uri, strictMatch) {
        return netutils.cookieMatchesURI.apply(netutils, arguments);
    }
};


NativeBarAPI.Autocomplete = function Autocomplete(componentInfo, logger) {
    this._componentInfo = componentInfo;
    this._logger = logger;
    this._providers = { __proto__: null };
};

NativeBarAPI.Autocomplete.prototype = {
    commonHistoryCategory: appCore.appName + "-history",
    searchComponentName: appCore.appName + "-autocomplete",
    
    
    addSearchProvider: function NativeAPI_Autocomplete_addSearchProvider(aSearchId, aProvider) {
        let ok = this._autocompleteSearchService.addSearchProvider(aSearchId, aProvider);
        if (!ok)
            throw new Error(strutils.formatString("Could not add search provider '%1'", [aSearchId]));
        this._providers[aSearchId] = aProvider;
    },
    
    
    removeSearchProvider: function NativeAPI_Autocomplete_removeSearchProvider(aSearchId) {
        let ok = this._autocompleteSearchService.removeSearchProvider(aSearchId, this._providers[aSearchId]);
        delete this._providers[aSearchId];
        if (!ok)
            throw new Error(strutils.formatString("Could not remove search provider '%1'", [aSearchId]))
    },
    _providers: null,
    
    get _autocompleteSearchService() {
        let serviceId = "@mozilla.org/autocomplete/search;1?name=" + this.searchComponentName;
        let autocompleteSearchService = Cc[serviceId].getService().wrappedJSObject;
        this.__defineGetter__("_autocompleteSearchService",
                              function NativeAPI_Autocomplete__autocompleteSearchService() autocompleteSearchService);
        return this._autocompleteSearchService;
    },
    
    _finalize: function NativeAPI_Autocomplete__finalize() {
        for (let [id, provider] in Iterator(this._providers)) {
            this._logger.warn("Removing dangling autocomplete provider " + id);
            this._autocompleteSearchService.removeSearchProvider(id, provider);
        }
    }
};



NativeBarAPI.DistrData = function DistrData() {
};

NativeBarAPI.DistrData.prototype = {
    
    getRecord: function NativeAPI_DistrData_getRecord(id) {
        let oldCore = application.core.oldCore;
        return (oldCore && oldCore.vendorData[id]) || null;
    }
};
