"use strict";

const EXPORTED_SYMBOLS = ["BarPlatform"];
const PLATFORM_SCRIPTS = [
    "preset.js", "manifest.js", "preset-with-manifest.js", "package.js", "unit.js", "permissions.js", "blacklist.js"
];

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;
const GLOBAL = this;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

var barApp;

const BarPlatform = {
    init: function BarPlatform_init(barApplication) {
        barApp = this._application = barApplication;
        barApplication.core.Lib.sysutils.copyProperties(barApplication.core.Lib, GLOBAL);
        
        this._loggersRoot = barApplication.name + ".CP";
        this._logger = Log4Moz.repository.getLogger(this._loggersRoot);
        this._loadModules();
    },
    
    finalize: function BarPlatform_finalize() {
        this._parsers = null;
        this._logger = null;
    },
    
    registerUnitParser: function BarPlatform_registerUnitParser(platformType, componentType, parser) {
        let parserKey = platformType + "#" + componentType;
        if (!parser)
            delete this._parsers[parserKey];
        else
            this._parsers[parserKey] = parser;
    },
    
    
    findPlatformMatch: function BarPlatform_findPlatformMatch(items) {
        if (!sysutils.isArray(items))
            throw new CustomErrors.EArgRange("items", "Array", items);
        
        let platformInfo = sysutils.platformInfo;
        for (let i = 0, len = items.length; i < len; i++) {
            let item = items[i];
            if (
                (("os" in item) && (item.os != platformInfo.os.name)) ||
                (("browser" in item) && (item.browser != platformInfo.browser.name)) ||
                (("architecture" in item) && (item.architecture != platformInfo.browser.simpleArchitecture))
            )
                continue;
            return item;
        }
        return null;
    },
    
    
    findBestLocalizedElement: function BarPlatform_findBestLocalizedElement(elementsArray) {
        let localizedElements = {};
        for each (let element in elementsArray) {
            let elemLocaleStr = element.getAttribute("locale") || element.getAttribute("lang") || "";
            localizedElements[elemLocaleStr] = element;
        }
        
        return misc.findBestLocalizedValue(localizedElements, this._application.locale);
    },

    parseComponentID: function BarPlatform_parseComponentID(componentID) {
        if (typeof componentID != "string")
            throw new CustomErrors.EArgType("componentID", "string", componentID);
        
        let separatorPos = componentID.indexOf("#");
        let packageID = componentID.substring(0, separatorPos);
        if (!packageID)
            throw new Error(this._consts.ERR_INVALID_COMP_ID + componentID);
        let compName = componentID.substring(separatorPos + 1);
        if (!compName)
            throw new Error(this._consts.ERR_INVALID_COMP_ID + componentID);
        return [packageID, compName];
    },
    
    getNewWidgetInstanceID: function BarPlatform_getNewWidgetInstanceID() {
        return "" + Date.now() + this._newWID++;
    },
    
    parsePrefPath: function BarPlatform_parsePrefPath(prefPath, compID) {
        let appCore = this._application.core;
        let [packageID, compName] = this.parseComponentID(compID);
        let partsRE = new RegExp( strutils.escapeRE(appCore.nativesPrefsPath + packageID) + "(\#(.+)\\.(\\w+))?\\.settings\\.(.+)");
        let partsMatch = prefPath.match(partsRE);
        if (!partsMatch)
            throw new Error("Invalid prefPath");
        let isPackagePref = !partsMatch[1],
            scope = partsMatch[3],
            isInstancePref = (!isPackagePref && (scope != "all")),
            settingName = partsMatch[4];
        let prefProps = {
            settingName: settingName,
            isPackagePref: isPackagePref,
            isComponentPref: !(isPackagePref || isInstancePref),
            isInstancePref: isInstancePref,
            instanceID: isInstancePref ? scope : undefined
        };
        return prefProps;
    },
    
    eraseSettings: function BarPlatform_eraseSettings() {
        let barCore = this._application.core;
        this._obsService.notifyObservers(null, barCore.eventTopics.EVT_BEFORE_GLOBAL_RESET, null);
        try {
            Preferences.resetBranch(barCore.xbWidgetsPrefsPath);
            Preferences.resetBranch(barCore.nativesPrefsPath);
            
            fileutils.removeFileSafe(this._application.directories.nativeStorageDir);
            let staticPrefsInfo = JSON.parse( fileutils.readStringFromStream(
                    this._application.addonFS.getStream(this._consts.STR_STATIC_PREFS_PATH) ) );
            const APP_NAME = this._application.name;
            for each (let [, prefPath] in Iterator(staticPrefsInfo.preferences))
                Preferences.reset(APP_NAME + prefPath);
            for each (let [, branchPath] in Iterator(staticPrefsInfo.branches))
                Preferences.resetBranch(APP_NAME + branchPath);
        }
        finally {
            this._obsService.notifyObservers(null, barCore.eventTopics.EVT_AFTER_GLOBAL_RESET, null);
        }
    },
    
    _parsers: {},
    _application: null,
    _loggersRoot: undefined,
    _logger: null,
    _newWID: 0,
    
    get _obsService() {
        let obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        delete this._obsService;
        return this._obsService = obs;
    },
    _consts: {
        STR_STATIC_PREFS_PATH: "$custombar/preferences/static-widgets/resetprefs.json",
        ERR_INVALID_COMP_ID: "Invalid component ID "
    },
    
    _loadModules: function BarPlatform__loadModules() {
        const mozSSLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
        const xbDirPath = this._application.partsURL + "platform/";
        PLATFORM_SCRIPTS.forEach(function BarPlatform_loadModule(scriptFileName) {
            this._logger.debug("  Including script " + scriptFileName);
            mozSSLoader.loadSubScript(xbDirPath + scriptFileName);
        }, this);
    },
    
    _getLogger: function BarPlatform__getLogger(name) {
        return Log4Moz.repository.getLogger(this._loggersRoot + "." + name);
    },
    
    _getParser: function BarPlatform__getParser(platformType, componentType) {
        let parser = this._parsers[platformType + "#" + componentType];
        if (!parser)
            throw new Error(strutils.formatString("No \"%2\" parser registered for API \"%1\".",
                                                  [platformType, componentType]));
        return parser;
    }
};

BarPlatform.__defineGetter__("IWidgetPrototype", function () {
    let intf = new sysutils.Interface("IWidgetPrototype", [
        
        "createInstance",
        "finalize"
    ], {
        id: "string",
        name: "string",
        isUnique: "boolean",
        unit: BarPlatform.Unit,
        pkg: BarPlatform.ComponentPackage,
        iconPath: "string",
        iconURI: "string",
        spawnedIDs: undefined,
        widgetSettings: undefined,
        packageSettings: undefined
    });
    delete BarPlatform.IWidgetPrototype;
    BarPlatform.__defineGetter__("IWidgetPrototype", function () intf);
    return BarPlatform.IWidgetPrototype;
});

BarPlatform.__defineGetter__("IWidgetInstance", function () {
    let intf = new sysutils.Interface("IWidgetInstance", [
        
        "buildUI",
        
        "getSettingValue",
        
        "applySetting",
        
        "applySettings",
        
        "eraseSettings",
        
        "finalize"
    ], {
        id: "string",
        prototype: BarPlatform.IWidgetPrototype,
        host: "object",
        instanceSettings: undefined
    });
    delete BarPlatform.IWidgetInstance;
    BarPlatform.__defineGetter__("IWidgetInstance", function () intf);
    return BarPlatform.IWidgetInstance;
});


