"use strict";

const EXPORTED_SYMBOLS = ["statLogger"];

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;

var Lib;
const statLogger = {
    init: function statLogger_init(aApplication) {
        this._application = aApplication;
        this._logger = aApplication.getLogger("StatLog");
        Lib = this._application.core.Lib;
        this.clear();
        
        this._database = new Lib.misc.Database(this._cacheFile, this._INIT_TABLE_QUERY);
    },
    
    finalize: function statLogger_finalize(doCleanup) {
        if (this._database) {
            this._database.close( doCleanup ?
                    Lib.fileutils.removeFileSafe.bind(Lib.fileutils, this._cacheFile) :
                    undefined );
        }
        this._database = null;
    },
    dump: function statLogger_dump() {
        this._logger.debug("Statistics:" +
            "\nshort actions:\n" + Lib.sysutils.dump(this._shortActions, 2) +
            "\nsys actions:\n" + Lib.sysutils.dump(this._sysActions, 2) +
            "\ncustom actions:\n" + Lib.sysutils.dump(this._customActions, 2));
    },
    
    clear: function statLogger_clear() {
        this._shortActions = {};
        this._sysActions = {};
        this._customActions = {};
    },
    
    logCustomAction: function statLogger_logCustomAction(componentID, actionID) {
        this._incAction(this._customActions, componentID, actionID);
    },
    
    logSysAction: function statLogger_logSysAction(componentID, actionID) {
        this._incAction(this._sysActions, componentID, actionID);
    },
    
    logShortAction: function statLogger_logShortAction(actionID) {
        if (!(actionID = this._filterActionID(actionID)))
            return;
        
        if (!(actionID in this._shortActions))
            this._shortActions[actionID] = 0;
        
        this._shortActions[actionID]++;
    },
    
    readActions: function statLogger_readActions() {
        let allChunks = [];
        
        for (let actionID in this._shortActions) {
            let actionCount = this._shortActions[actionID];
            let chunk = actionID + ((actionCount > 1) ? ("." + actionCount) : "");
            allChunks.push(chunk);
        }
        
        for (let componentHash in this._sysActions) {
            for (let actionID in this._sysActions[componentHash]) {
                let actionCount = this._sysActions[componentHash][actionID];
                let chunk = componentHash + "-" + actionID + ((actionCount > 1) ? ("." + actionCount) : "");
                allChunks.push(chunk);
            }
        }
        
        for (let componentHash in this._customActions) {
            for (let actionID in this._customActions[componentHash]) {
                let actionCount = this._customActions[componentHash][actionID];
                let chunk = componentHash + "+" + actionID + ((actionCount > 1) ? ("." + actionCount) : "");
                allChunks.push(chunk);
            }
        }
        
        this.clear();
        
        return allChunks.join(",");
    },
    
    ACTIONS: {
        COMP_INSTALL: 1,
        COMP_REMOVE: 2,
        COMP_DAYUSE: 3,
        DP_HASH: 4,
        
        BTN_CLICK: 10,
        MENU_CLICK: 11
    },
    logAddonEvents: function statLogger_logAddonEvents(events ) {
        this._database.execQuery("BEGIN TRANSACTION");
        for (let [key, value] in Iterator(events)) {
            this._database.execQuery("INSERT OR REPLACE INTO addonevents (key, value) VALUES (:key, :value)", {
                key: key,
                value: "" + value
            });
        }
        this._database.execQuery("COMMIT TRANSACTION");
    },
    
    
    fetchAddonEvents: function statLogger_fetchAddonEvents() {
        let fetchQuery = "SELECT * FROM addonevents";
        if (this._lastReturnedRowID >= 0)
            fetchQuery += " WHERE rowid > " + this._lastReturnedRowID;
        
        let rows = this._database.execQuery(fetchQuery);
        let firstRow = rows[0];
        let lastRow = firstRow && this._database.execQuery("SELECT rowid FROM addonevents ORDER BY rowid DESC LIMIT 1")[0];
        
        let firstRowID = firstRow && (this._lastReturnedRowID + 1);
        let lastRowID = lastRow && lastRow.rowid;
        let rowRange = this._EMPTY_ROW_RANGE;
        
        if (rows.length > 0) {
            rowRange = [firstRowID, lastRowID];
            this._lastReturnedRowID = lastRowID;
        }
        
        return [rows, rowRange];
    },
    
    
    eraseAddonEvents: function statLogger_eraseAddonEvents(rowRange) {
        if (rowRange === this._EMPTY_ROW_RANGE)
            return;
        
        let [firstRowID, lastRowID] = rowRange;
        if (!Lib.sysutils.isNumber(firstRowID) || !Lib.sysutils.isNumber(lastRowID))
            throw new TypeError("Invalid transaction row range. rowRange: " + Lib.sysutils.dump(rowRange));
        
        this._database.execQuery("DELETE FROM addonevents WHERE rowid >= :firstRowID  AND rowid <= :lastRowID",
                                 {firstRowID: firstRowID, lastRowID: lastRowID});
        
        this._updateLastRow();
    },
    clearAddonEvents: function statLogger_clearAddonEvents() {
        this._database.execQuery("DELETE FROM addonevents");
        this._updateLastRow();
    },
    
    _sysActions: null,
    _customActions: null,
    _shortActions: null,
    
    _database: null,
    _lastReturnedRowID: 0,
    _EMPTY_ROW_RANGE: null,
    _INIT_TABLE_QUERY: "CREATE TABLE IF NOT EXISTS addonevents (key TEXT, value BLOB)",
    
    get _cacheFile() {
        let cacheFile = this._application.directories.appRootDir;
        cacheFile.append("statcache.sqlite");
        return cacheFile;
    },
    
    _updateLastRow: function statLogger__updateLastRow() {
        let lastRow = this._database.execQuery("SELECT rowid FROM addonevents ORDER BY rowid DESC LIMIT 1")[0];
        let lastStoredRowID = lastRow? lastRow.rowid : 0;
        if (lastStoredRowID < this._lastReturnedRowID)
            this._lastReturnedRowID = lastStoredRowID;
    },
    _filterActionID: function statLogger__filterActionID(actionID) {
        let action = Number(actionID);
        
        if (action > 0) {
            action = "" + action;
            if (action === ("" + actionID))
                return action;
        }
        
        return null;
    },
    
    _incAction: function statLogger__incAction(actionsMap, componentID, actionID) {
        if (!(actionID = this._filterActionID(actionID)))
            return;
        
        let componentHash = this._compID2Hash(componentID);
        let compActions = actionsMap[componentHash] || (actionsMap[componentHash] = {});
        
        if (!(actionID in compActions))
            compActions[actionID] = 0;
        
        compActions[actionID]++;
    },
    
    _compID2Hash: function statLogger__incAction(componentID) {
        return Lib.misc.CryptoHash.getFromString(componentID, "MD5");
    }
};
