"use strict";

XB.cache = {
    _database: null,
    
    init: function XBCache_init() {
        this._logger = XB._base.getLogger("Cache");
        this._database = new misc.Database(this._cacheFile, this._consts.INIT_TABLE_QUERY);
        this.eraseOldRecords();
    },
    
    finalize: function XBCache_finalize(doCleanup) {
        try {
            if (this._database) {
                this._database.close( doCleanup ?
                        fileutils.removeFileSafe.bind(fileutils, this._cacheFile) :
                        undefined );
            }
        }
        catch (e) {
            this._logger.error("Failed closing database. " + e.message);
        }
        
        this._database = null;
        this._logger = null;
    },
    
    
    store: function XBCache_store(key, value) {
        if (sysutils.isObject(value))
            value = JSON.stringify(value);
        let storeTime = this._currentTimestampSecs;
        this._logger.debug("Storing data with key " + key + " at time " + storeTime);
        this._database.execQueryAsync("INSERT OR REPLACE INTO resources (id, time, value) VALUES (:id, :time, :value)", {
            id: key,
            time: storeTime,
            value: value.toString()
        });
    },
    
    startSearch: function XBCache_startSearch(key, notOlderThan, callback) {
        let treshold = notOlderThan? (this._currentTimestampSecs - notOlderThan): 0;
        this._logger.debug("Started looking for data " + key + " not older than " + (new Date(treshold * 1000)) );
        let dbCallback = this._onSearchComplete.bind(this, callback);
        
        return this._database.execQueryAsync(this._consts.SEARCH_QUERY, {id: key, treshold: treshold}, dbCallback);
    },
    
    eraseRecord: function XBCache_eraseRecord(key) {
        this._logger.debug("Deleting cache record with key " + key);
        this._database.execQueryAsync("DELETE FROM resources WHERE (id = :id)", {id: key});
    },
    
    
    eraseOldRecords: function XBCache_eraseOldRecords() {
        this._database.execQueryAsync("DELETE FROM resources WHERE (time < :time)", {
            time: (this._currentTimestampSecs - 3600 * 24 * 90)
        });
    },
    
    
    flush: function XBCache_flush() {
        this._database.execQueryAsync("DELETE FROM resources");
    },
    
    _consts: {
        INIT_TABLE_QUERY: "CREATE TABLE IF NOT EXISTS resources (id TEXT, time INTEGER, value BLOB, PRIMARY KEY(id))",
        SEARCH_QUERY: "SELECT * FROM resources WHERE id = :id AND time >= :treshold LIMIT 1"
    },
    
    
    get _currentTimestampSecs() {
        return parseInt(Date.now() / 1000, 10);
    },
    
    get _cacheFile() {
        let cacheFile = XB._base.application.directories.appRootDir;
        cacheFile.append("xbcache.sqlite");
        return cacheFile;
    },
    
    _onSearchComplete: function XBCache__onSearchComplete(userCallback, rows, storageError) {
        let cachedData;
        let cacheTimestamp;
        if (!storageError) {
            let row = rows[0];
            if (row) {
                cacheTimestamp = row.time;
                cachedData = row.value;
            }
        }
        userCallback(cachedData, cacheTimestamp, storageError);
    }
};
