"use strict";


XB.types = {
    
    isXML: function XBRuntime_isXML(xVal) {
        return (xVal instanceof this.XML);
    },
    
    isException: function XBRuntime_isException(xVal) {
        return xVal instanceof this.Exception;
    },
    
    isRequestData: function XBRuntime_isRequestData(xVal) {
        return xVal instanceof this.RequestData;
    },
    ensureValueTypeIs: function XBTypes_ensureValueTypeIs(xValue, xType) {
        let expectedRank = this._ranksMap[xType];
        if (expectedRank === undefined)
            throw new Error(this._consts.ERR_UNKNOWN_TYPE + xType);
        let valueRank = this.rankValue(xValue);
        if (expectedRank != valueRank)
            throw new TypeError(xType + " expected, got " + this._typeNames[valueRank]);
    },
    describeValue: function XBRuntime_describeValue(xValue) {
        let valueRank = this.rankValue(xValue);
        let descr = this._typeNames[valueRank];
        if (valueRank > this._ranksMap.Empty && valueRank < this._ranksMap.XML)
            descr += "(" + this.xToString(xValue) + ")";
        if (valueRank == this._ranksMap.Exception)
            descr += "(" + xValue.toString() + ")";
        return descr;
    },
    
    rankValue: function XBRuntime__rankValue(value) {
        if (value === XB.types.empty)
            return this._ranksMap.Empty;
        if (sysutils.isBool(value))
            return this._ranksMap.Bool;
        if (sysutils.isNumber(value))
            return this._ranksMap.Number;
        if (sysutils.isString(value))
            return this._ranksMap.String;
        if (this.isXML(value))
            return this._ranksMap.XML;
        if (this.isRequestData(value))
            return this._ranksMap.RequestData;
        if (this.isException(value))
            return this._ranksMap.Exception;
        
        throw new TypeError(this._consts.ERR_UNKNOWN_TYPE + typeof value);
    },
    
    xToString: function XBRuntime_xToString(fromXVal) {
        if (sysutils.isString(fromXVal))
            return fromXVal;
        if (fromXVal == undefined || fromXVal == XB.types.empty)
            return "";
        if (this.isXML(fromXVal))
            return fromXVal.textContent;
        if (this.isRequestData(fromXVal))
            return fromXVal.url;
        if (this.isException(fromXVal))
            throw new this.ETypeCast(this.describeValue(fromXVal), "String");
        return fromXVal.toString();
    },
    
    xToNumber: function XBRuntime_xToNumber(fromXVal) {
        if (sysutils.isNumber(fromXVal)) {
            if (isNaN(fromXVal))
                throw new this.ETypeCast(this.describeValue(fromXVal), "Number");
            return fromXVal;
        }
        if (this.isException(fromXVal))
            throw new this.ETypeCast(this.describeValue(fromXVal), "Number");
        if (sysutils.isBool(fromXVal))
            return Number(fromXVal);
        let trimmedStrVal = this.xToString(fromXVal).trim();
        if (!trimmedStrVal)
            return 0;
        if (!this._numPattern.test(trimmedStrVal))
            throw new this.ETypeCast(this.describeValue(fromXVal), "Number");
        let num = parseFloat(trimmedStrVal);
        if (isNaN(num))
            throw new this.ETypeCast(this.describeValue(fromXVal), "Number");
        return num;
    },
    
    xToBool: function XBRuntime_xToBool(fromXVal) {
        if (sysutils.isBool(fromXVal))
            return fromXVal;
        if (sysutils.isNumber(fromXVal))
            return fromXVal != 0;
        if (this.isException(fromXVal))
            throw new this.ETypeCast(this.describeValue(fromXVal), "Bool");
        let strVal = strutils.trimSpaces( this.xToString(fromXVal) );
        if (strVal == "" || strVal == "false")
            return false;
        if (this._numPattern.test(strVal)) {
            let numVal = Number(strVal);
            if (!isNaN(numVal))
                return numVal != 0;
        }
        return true;
    },
    
    xToXML: function XBRuntime_xToXML(fromXVal, destDoc) {
        if (this.isXML(fromXVal))
            return fromXVal;
        return new XB.types.XML(destDoc, destDoc.createTextNode(this.xToString(fromXVal)));
    },
    
    xToRequestData: function XBRuntime_xToRequestData(fromXVal) {
        if (this.isRequestData(fromXVal))
            return fromXVal;
        let monthTimeRange = this._consts.SECONDS_IN_MONTH;
        return new XB.types.RequestData(this.xToString(fromXVal), "GET", monthTimeRange, monthTimeRange * 2,
                                        XB.types.RequestData.Format.FMT_XML, {start:100, end:399});
    },
    
    
    makeXML: function XBRuntime_makeXML(xmlJson, destDoc ) {
        const DOMImpl = this._DOM_IMPL;
        function makeElement(elementJson, asDoc) {
            let element;
            if (!asDoc)
                element = destDoc.createElement(elementJson.name);
            else {
                destDoc = DOMImpl.createDocument("", elementJson.name, null);
                element = destDoc.documentElement;
            }
            
            for (let attrName in elementJson.attributes)
                element.setAttribute(attrName, elementJson.attributes[attrName]);
            makeXML_append(elementJson.children, element);
            return asDoc ? destDoc : element;
        }
        
        function makeXML_append(children, parent) {
            for each (let elementJson in children)
                parent.appendChild(makeElement(elementJson));
        }
        
        if (!sysutils.isArray(xmlJson))
            return new XB.types.XML(makeElement(xmlJson, true));
        
        let result = new XB.types.XML(destDoc, []);
        makeXML_append(xmlJson, result);
        return result;
    },
    createXBExceptionFromRTError: function XBRuntime_XBEFromRTE(nid, error) {
        let XBE = new XB.types.Exception(nid, XB.types.Exception.types.E_RUNTIME, error.message);
        XBE.name = error.name;
        XBE.fileName = error.fileName;
        XBE.lineNumber = error.lineNumber;
        XBE.original = error;
        
        return XBE;
    },
    
    
    compareValues: function XBRuntime_compareValues(left, right, mode, bufferXMLDoc, castFails) {
        if (left === right)
            return 0;
        let commonRank = this._findCommonRank(left, right, mode, castFails);
        try {
            var value1 = this._castToRank(left, commonRank, bufferXMLDoc),
                value2 = this._castToRank(right, commonRank, bufferXMLDoc);
        }
        catch (e) {
            if (castFails)
                throw e;
            commonRank = NaN;
        }
        
        let ranksMap = this._ranksMap;
        switch (commonRank) {
            case ranksMap.Empty:
                return 0;
            case ranksMap.Bool:
                return Number(value1) - Number(value2);
            case ranksMap.Number:
                return value1 - value2;
            case ranksMap.String:
                return strutils.compareStrings(value1, value2);
            case ranksMap.XML:
            case ranksMap.RequestData:
            case ranksMap.Exception:
                return value1.equalsTo(value2) ? 0 : NaN;
            default:
                return NaN;
        }
    },
    cmpModes: {
        CMP_STRICT: 0,
        CMP_FREE: 1,
        CMP_SMART: 2
    },
    
    _consts: {
        ERR_XPATH_QUERY_FAILED: "XPATH query '%1' failed. %2",
        ERR_UNKNOWN_TYPE: "Unknown type ",
        ERR_NO_CAST_FUNC: "No casting function for this rank",
        SECONDS_IN_MONTH: 60 * 60 * 24 * 30
    },
    _DOM_IMPL: Cc["@mozilla.org/xml/xml-document;1"].createInstance(Ci.nsIDOMDocument).implementation,
    
    _numPattern: /^[+\-]?(\d+\.?|\d*\.\d+)$/,
    
    _typeNames: {
        0: "Empty",
        1: "Bool",
        2: "Number",
        3: "String",
        4: "XML",
        5: "RequestData",
        6: "Exception"
    },
    
    _ranksMap: {
        Empty: 0,
        Bool: 1,
        Number: 2,
        String: 3,
        XML: 4,
        RequestData: 5,
        Exception: 6
    },
    
    _castToRank: function XBRuntime__castToRank(value, rank, bufferXMLDoc) {
        let destType = this._typeNames[rank];
        let castFunc = this["xTo" + destType];
        if (castFunc == undefined)
            throw new Error(this._consts.ERR_NO_CAST_FUNC);
        return castFunc.call(this, value, bufferXMLDoc);
    },
    
    _findCommonRank: function XBRuntime__findCommonRank(value1, value2, mode, castFails) {
        let r1 = this.rankValue(value1);
        let r2 = this.rankValue(value2);
        
        if (r1 == r2)
            return r1;
        
        if (mode == this.cmpModes.CMP_STRICT ||
            r1 > this._ranksMap.RequestData || r2 > this._ranksMap.RequestData)
            return NaN;
        
        let maxRank = Math.max(r1, r2);
        let commonRank = (mode == this.cmpModes.CMP_SMART ? Math.min(this._ranksMap.String, maxRank) : maxRank);
        return commonRank;
    }
};

XB.types.EXPATH = CustomErrors.ECustom.extend({
    $name: "EXPATH",
    
    constructor: function XBRuntimeEXPATH(expression) {
        this.base();
        this._expr = expression.toString();
    },
    
    _message: "XPATH query error",
    _expr: undefined,
    
    get _details() {
        return [this._expr];
    }
});

XB.types.EXPATHNaN = XB.types.EXPATH.extend({
    $name: "EXPATHNaN",
    
    _message: "XPATH 'number' expression returned NaN"
});

XB.types.EXPATHUnexpectedResult = XB.types.EXPATH.extend({
    $name: "EXPATHUnexpectedResult",
    
    constructor: function XBRuntimeEXPATHUnexpReslt(expression, resultType) {
        this.base(expression);
        this._resType = resultType.toString();
    },
    
    _message: "XPATH query error",
    _resType: undefined,
    
    get _details() {
        return [this._expr, this._resType];
    }
});

XB.types.ETypeCast = CustomErrors.ECustom.extend({
    $name: "ETypeCast",
    
    constructor: function XBRuntimeETypeCast(origValueDescr, destTypeName) {
        this.base();
        this._origValueDescr = origValueDescr.toString();
        this._destTypeName = destTypeName.toString();
    },
    
    _message: "This value conversion is impossible",
    _origValueDescr: undefined,
    _destTypeName: undefined,
    
    get _details() {
        return [this._origValueDescr, this._destTypeName];
    }
});


XB.types.empty = null;


XB.types.XML = function XBXML(bufferDoc , _ ) {
    if (!(bufferDoc instanceof Ci.nsIDOMDocument))
        throw new CustomErrors.EArgType("bufferDoc", "nsIDOMDocument", bufferDoc);
    this._bufferDoc = bufferDoc;
    
    if (arguments.length == 1) {
        this._isADocument = true;
        this._rootNode = bufferDoc.documentElement;
    }
    else {
        this._initAsList();
        this.appendNodes.apply(this, Array.slice(arguments, 1));
    }
};
XB.types.XML.persistentProperties = ["_valType", "_rootNode"];

XB.types.XML.prototype = {
    constructor: XB.types.XML,
    
    
    appendNodes: function XML_appendNodes() {
        if (this._isADocument) {
            let oldRoot = this._rootNode;
            this._initAsList();
            this._rootNode.appendChild(oldRoot);
        }
        
        for (let argIndex = 0, argsLen = arguments.length; argIndex < argsLen; argIndex++) {
            let arg = arguments[argIndex];
            
            if (sysutils.isArray(arg)) {
                this.appendNodes.apply(this, arg);
                continue;
            }
            
            if (arg instanceof Ci.nsIDOMNode) {
                this._appendAnyNode(arg);
                continue;
            }
            
            let isThisType = (arg instanceof this.constructor);
            if (arg instanceof Ci.nsIDOMNodeList || isThisType) {
                for (let nodeIndex = 0, nodesLen = arg.length; nodeIndex < nodesLen; nodeIndex++) {
                    this._appendAnyNode(arg.item(nodeIndex));
                }
                if (isThisType) {
                    let argAttrs = arg.attributes;
                    for (let i = 0, len = argAttrs.length; i < len; i++) {
                        this._appendAnyNode(argAttrs.item(i));
                    }
                }
                continue;
            }
            
            throw new TypeError(XB._base.consts.ERR_DOM_NODES_EXPECTED);
        }
    },
    
    addAttribute: function XML_addAttribute(ns, name, value) {
        if (this.disposed)
            throw new Error(XB._base.consts.ERR_DISPOSED_XML);
        this._rootNode.setAttributeNS(ns, name, value);
    },
    
    toString: function XML_toString() {
        let role = this._isADocument? "document": "nodes";
        if (this.disposed)
            return "[XBXML " + role + " (disposed)]";
        return "[XBXML "+ role + " length: " + this.length + ", content:\n" +
            strutils.trimSpaces(this.serialize().substr(0, 800)) + "...\n]";
    },
    appendChild: function XML_appendChild(node) {
        this._rootNode.appendChild(node);
    },
    
    item: function XML_item(at) {
        if (!this._isADocument)
            return this._rootNode.childNodes[at];
        if (at > 0)
            throw new RangeError("Document has only one root node");
        return this._rootNode;
    },
    
    get attributes() {
        if (this.disposed)
            throw new Error(XB._base.consts.ERR_DISPOSED_XML);
        return this._rootNode.attributes;
    },
    
    get childNodes() {
        return this._rootNode.childNodes;
    },
    
    get length() {
        if (this.disposed)
            return 0;
        if (this._isADocument)
            return 1;
        return this._rootNode.childNodes.length;
    },
    
    get disposed() {
        return !this._rootNode;
    },
    dispose: function XML_dispose() {
        if (this._isADocument) {
            this._isADocument = false;
        }
        else {
            if (this._frag)
                this._frag.removeChild(this._rootNode);
            this._frag = null;
        }
        
        if (this.disposed)
            XB._base.logger.warn("Attemp to dispose an already disposed XML object.");
        else {
            if (XB._base.logger.level <= Log4Moz.Level.Trace)
                XB._base.logger.trace("Disposing XML:\n" + this.serialize().substr(0, 400) + "\n...");
            this._rootNode = null;
        }
    },
    
    serialize: function XML_serialize() {
        let result = "";
        for (let nodeIndex = 0, len = this.length; nodeIndex < len; nodeIndex++) {
            result += xmlutils.serializeXML(this.item(nodeIndex));
        }
        return result;
    },
    
    get textContent() {
        return this._rootNode.textContent;
    },
    
    equalsTo: function XML_equalsTo(other) {
        if ( !(other instanceof this.constructor) )
            throw new TypeError("XML type expected");
        if (other === this)
            return true;
        if ( (this._isADocument ^ other._isADocument) )
            return false;
        
        if (this.disposed || other.disposed)
            return false;
        return this._rootNode.isEqualNode(other._rootNode);
    },
    
    query: function XML_query(expr, NSResolver) {
        if (this.disposed)
            throw new Error("Querying from disposed XML object (expression: \"" + expr + "\")");
        
        if (XB._base.logger.level <= Log4Moz.Level.Trace) {
            XB._base.logger.trace("Query from:\n" + this.serialize().substr(0, 400) + "\n...");
        }
        let queryNode = this._isADocument? this._rootNode.ownerDocument: this._rootNode;
        return this._processQueryResult( this._queryXMLDocCheck(expr, queryNode, NSResolver) );
    },
    
    transform: function XML_transform(stylesheet, destDoc) {
        let src = (this._isADocument) ? this._rootNode.ownerDocument : this._rootNode;
        let newXML = new XB.types.XML(destDoc, xmlutils.transformXMLToFragment(src, stylesheet, destDoc));
        return newXML;
    },
    
    clone: function XML_clone() {
        return new this.constructor(this);
    },
    
    get owner() {
        return this._owner;
    },
    set owner(owner) {
        this._owner = owner;
    },
    
    setRoot: function XBXML_setRoot(rootNode) {
        this._rootNode = this._frag.appendChild(rootNode);
        if (!this._rootNode)
            throw new Error("Couldn't set root node");
        this._isADocument = false;
    },
    
    _bufferDoc: null,
    _valType: "XBXML",
    _isADocument: false,
    _frag: null,
    _rootNode: null,
    _owner: null,
    
    _initAsList: function XML__initAsList() {
        this._frag = this._bufferDoc.createDocumentFragment();
        this._rootNode = this._frag.appendChild(this._bufferDoc.createElement("xml"));
        if (!this._rootNode)
            throw new Error("Couldn't create root node");
        this._isADocument = false;
    },
    
    _appendAnyNode: function XML__appendAnyNode(node) {
        let newNode = null;
        if (node.ownerDocument == this._rootNode.ownerDocument) {
            newNode = node.cloneNode(true);
        } else {
            try {
                newNode = this._rootNode.ownerDocument.importNode(node, true);
            }
            catch(e) {
                XB._base.logger.warn("Native importNode failed. " + e);
                newNode = xmlutils.recreateXML(node, this._rootNode.ownerDocument, true);
            }
        }
        
        if (newNode.nodeType == newNode.ATTRIBUTE_NODE)
            this._rootNode.setAttributeNS(newNode.namespaceURI, newNode.localName, newNode.value);
        else
            this._rootNode.appendChild(newNode);
    },
    
    
    _queryXMLDocCheck: function XBXML_queryXMLDocCheck(expr, contextNode, extNSResolver) {
        try {
            let result = xmlutils.queryXMLDoc(expr, contextNode, extNSResolver);
            if ((typeof result == "number") && isNaN(result))
                throw new XB.types.EXPATHNaN(expr);
            if (result === undefined)
                throw new XB.types.EXPATHUnexpectedResult(expr, typeof result);
            
            XB._base.logger.trace("XPATH (" + expr + ") result is \"" + result + "\"");
            return result;
        }
        catch (e) {
            if (!(e instanceof XB.types.EXPATHNaN))
                XB._base.logger.warn(strutils.formatString(XB.types._consts.ERR_XPATH_QUERY_FAILED,
                                                           [expr, strutils.formatError(e)]));
            throw e;
        }
    },
    
    _processQueryResult: function XBXML__processQueryResult(queryResult ) {
        if (sysutils.isArray(queryResult)) {
            XB._base.logger.trace("XML query returned an array of nodes " + queryResult.length);
            if (!queryResult.length)
                return XB.types.empty;
            
            let xmlContent = (queryResult.length == 1) ? queryResult[0] : queryResult;
            queryResult = (xmlContent instanceof Ci.nsIDOMDocument) ?
                    new XB.types.XML(xmlContent) :
                    new XB.types.XML(this._bufferDoc, xmlContent);
            
            XB._base.logger.trace("Constructed a new XML " + queryResult.length);
        }
        return queryResult;
    }
};


XB.types.Exception = function XBException(srcNodeUid, eType, msg) {
    this._type = eType || this._type;
    this._msg = msg;
    this._srcNodeUid = srcNodeUid;
};

XB.types.Exception.types = {
    E_GENERIC: "Exception",
    E_SYNTAX: "Syntax",
    E_TYPE: "Type",
    E_RUNTIME: "Runtime",
    E_SECURITY: "Security",
    E_RETHROW: "Rethrow",
    E_LASTVALUE: "LastValue"
};

XB.types.Exception.prototype = {
    get type() {
        return this._type;
    },
    
    get srcNodeUid() {
        return this._srcNodeUid;
    },
    
    get message() {
        return this._msg;
    },
    
    equalsTo: function XBException_equalsTo(other) {
        return sysutils.objectsAreEqual(this, other);
    },
    
    toString: function XBException_toString() {
        return this._type + "@" + this._srcNodeUid + ": " + this._msg;
    },
    
    _type: XB.types.Exception.types.E_GENERIC,
    _srcNodeUid: undefined
};

XB.types.RequestData = function XBRequestData(url, method, update, expire, format, validStatusRange, validXpath, specials, isPrivate) {
    let typeUtils = XB.types;
    
    typeUtils.ensureValueTypeIs(url, "String");
    let addrScheme = XB._base.nsIIOService.extractScheme(url);
    if (addrScheme !== "http" && addrScheme !== "https")
        throw new CustomErrors.ESecurityViolation("XB.types.RequestData", "Address scheme is " + addrScheme);
    this._url = url;
    
    typeUtils.ensureValueTypeIs(method, "String");
    this._method = method;
    
    typeUtils.ensureValueTypeIs(update, "Number");
    if (update < 1)
        throw new RangeError("Invalid update interval " + update);
    this._updateInterval = update;
    
    typeUtils.ensureValueTypeIs(expire, "Number");
    if (expire < 0)
        throw new RangeError("Invalid expiration time " + expire);
    this._expirationInterval = expire;
    
    typeUtils.ensureValueTypeIs(format, "Number");
    if (format < XB.types.RequestData.Format.FMT_TEXT || format > XB.types.RequestData.Format.FMT_JSON)
        throw new RangeError("Unknown format type");
    this._format = format;
    
    if (validStatusRange && sysutils.isNumber(validStatusRange.start) && sysutils.isNumber(validStatusRange.end) &&
        validStatusRange.start >= 100 && validStatusRange.end <= 599 && validStatusRange.start <= validStatusRange.end) {
        this._statusMin = validStatusRange.start;
        this._statusMax = validStatusRange.end;
    }
    else
        throw new TypeError("Invalid status range parameter");
    
    if (validXpath) {
        typeUtils.ensureValueTypeIs(validXpath, "String");
        this._checkXpathExpr = validXpath;
    }
    
    if (specials) {
        this._hiddenParams = specials.hidden;
    }
    
    this._isPrivate = !!isPrivate;
};

XB.types.RequestData.Format = {
    FMT_TEXT: 0,
    FMT_XML: 1,
    FMT_JSON: 2
};

XB.types.RequestData.prototype = {
    constructor: XB.types.RequestData,
    
    get url() {
        return this._url;
    },
    
    get method() {
        return this._method;
    },
    
    get isPrivate() {
        return this._isPrivate;
    },
    
    get updateInterval() {
        return this._updateInterval;
    },
    
    get expirationInterval() {
        return this._expirationInterval;
    },
    
    get format() {
        return this._format;
    },
    
    get statusRange() {
        return {start: this._statusMin, end: this._statusMax};
    },
    
    get xpathExpression() {
        return this._checkXpathExpr;
    },
    
    get hiddenParams() {
        return this._hiddenParams;
    },
    
    get specialsHash() {
        if (this._spHash)
            return this._spHash;
        
        let resultParts = [];
        for (let hp in this._hiddenParams) {
            resultParts.push(hp + ":" + this._hiddenParams[hp]);
        }
        if (!this._isPrivate)
            resultParts.push("pulic");
        
        return (this._spHash = resultParts.join("#"));
    },
    
    get hash() {
        return this._hash || (this._hash =
            [this._url, this._method, this.updateInterval, this.expirationInterval, this._format, this._statusMin,
             this._statusMax, this._checkXpathExpr, this.specialsHash].join("#"));
    },
    
    equalsTo: function ReqData_equalsTo(other) {
        if ( !(other instanceof this.constructor) )
            throw new TypeError("RequestData required");
        return this._fields.every(function(field) { return this[field] == other[field]; }, this);
    },
    
    _spHash: undefined,
    _hash: undefined,
    _url: undefined,
    _method: "GET",
    _updateInterval: 0,
    _expirationInterval: 0,
    _format: XB.types.RequestData.Format.FMT_TEXT,
    _statusMin: 0,
    _statusMax: 0,
    _checkXpathExpr: undefined,
    _hiddenParams: null,
    _isPrivate: null,
    
    _fields: ["_url", "_method", "_updateInterval", "_expirationInterval", "_format", "_statusMin", "_statusMax", "_checkXpathExpr", "specialsHash", "_isPrivate"]
};
