"use strict";

const EXPORTED_SYMBOLS = ["SimpleProtocol"];

const {
    classes: Cc,
    interfaces: Ci,
    results: Cr,
    utils: Cu,
    manager: Cm
} = Components;
const IOS = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
const InputStreamFabric = Cc["@mozilla.org/io/string-input-stream;1"];
const nsIStringInputStream = Ci.nsIStringInputStream;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

function SimpleProtocol(scheme) {
    this._scheme = "" + scheme;
    let uuidStr = Cc["@mozilla.org/uuid-generator;1"].getService(Ci.nsIUUIDGenerator).generateUUID().toString();
    this._classID = Components.ID(uuidStr);
    this._contractID = "@mozilla.org/network/protocol;1?name=" + scheme;
    this._dataProviders = [];
    
    Cm.QueryInterface(Ci.nsIComponentRegistrar).registerFactory(
        this._classID,
        this._DESCR,
        this._contractID,
        this
    );
}

SimpleProtocol.prototype = {
    constructor: SimpleProtocol,
    
    get wrappedJSObject() {
        return this;
    },
    
    unregister: function SimpleProtocol_unregister() {
        this._dataProviders = [];
        Cm.QueryInterface(Ci.nsIComponentRegistrar).unregisterFactory(this._classID, this);
    },
    
    addDataProvider: function SimpleProtocol_addDataProvider(provider) {
        if (this._dataProviders.indexOf(provider) > -1) return;
        
        if (!("newURI" in provider) && !("newChannel" in provider) && !("getContent" in provider))
            throw new Error("Bad data provider interface.");
        
        this._dataProviders.push(provider);
    },
    
    removeDataProvider: function SimpleProtocol_removeDataProvider(provider) {
        if (this._dataProviders)
            this._dataProviders = this._dataProviders.filter(function(handler) provider !== handler);
    },
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler]),
    
    createInstance: function SimpleProtocol_createInstance(aOuter , iid ) {
        return this.QueryInterface(iid);
    },
    
    lockFactory: function SimpleProtocol_lockFactory(lock) {},
    
    get scheme() {
        return this._scheme;
    },
    
    get protocolFlags() {
        return Ci.nsIProtocolHandler.URI_DANGEROUS_TO_LOAD;
    },
    
    get defaultPort() {
        return -1;
    },
    
    allowPort: function SimpleProtocol_allowPort(port, scheme) {
        return false;
    },
    
    newURI: function SimpleProtocol_newURI(spec, originalCharset, baseURI) {
        let simpleURI = this._makeSimpleURI(spec, baseURI);
        let result = this._findURI(spec, originalCharset, baseURI, simpleURI) || simpleURI;
        
        if (result)
            return result;
        return IOS.newURI("data:,", null, null);
    },
    
    newChannel: function SimpleProtocol_newChannel(uri) {
        try {
            let channel = this._findChannel(uri);
            
            if (!channel) {
                channel = IOS.newChannel("data:,", null, null);
                channel.originalURI = uri;
            }
            
            return channel;
        }
        catch (e) {
            Cu.reportError("SimpleProtocol(" + this._scheme + "): " + e + ". URI: " + uri.spec);
            throw Cr.NS_ERROR_FAILURE;
        }
    },
    
    _scheme: undefined,
    _classID: undefined,
    _contractID: undefined,
    _DESCR: "Simple protocol handler",
    _dataProviders: null,
    
    _findURI: function SimpleProtocol__findURI(spec, originalCharset, baseURI, simpleURI) {
        for each (let dataProvider in this._dataProviders) {
            if (typeof dataProvider.newURI != "function") continue;
            try {
                let uri = dataProvider.newURI(spec, originalCharset, baseURI, simpleURI);
                if (uri)
                    return uri;
            }
            catch (e) {
                Cu.reportError("SimpleProtocol(" + this._scheme + ") _findURI: " + e + ". Spec: " + spec);
            }
        }
        return null;
    },
    
    _makeSimpleURI: function SimpleProtocol__makeSimpleURI(spec, baseURI) {
        try {
            let simpleURI = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
            simpleURI.spec = baseURI ? baseURI.resolve(spec) : spec;
            return simpleURI;
        }
        catch (e) {
            return null;
        }
    },
    
    _findChannel: function SimpleProtocol__findChannel(uri) {
        let channel = null;
        
        for each (let dataProvider in this._dataProviders) {
            try {
                if (typeof dataProvider.newChannel == "function") {
                    channel = dataProvider.newChannel(uri);
                }
                else if (typeof dataProvider.getContent == "function") {
                    let strData = "" + dataProvider.getContent(uri);
                    let inputStream = InputStreamFabric.createInstance(nsIStringInputStream);
                    inputStream.setData(strData, strData.length);
                    
                    channel = Cc["@mozilla.org/network/input-stream-channel;1"]
                                .createInstance(Ci.nsIInputStreamChannel)
                                .QueryInterface(Ci.nsIChannel);
                    channel.setURI(uri);
                    channel.originalURI = uri;
                    channel.contentStream = inputStream;
                }
                
                if (channel !== null)
                    return channel;
            }
            catch (e) {
                Cu.reportError("SimpleProtocol(" + this._scheme + ") _findChannel: " + e + ". Spec: " + uri.spec);
            }
        }
        
        throw Cr.NS_ERROR_FAILURE;
    }
};
