"use strict";

const misc = {
    
    getBrowserWindows: function misc_getBrowserWindows() {
        let windows = [];
        let wndEnum = Cc["@mozilla.org/appshell/window-mediator;1"]
                        .getService(Ci.nsIWindowMediator)
                        .getEnumerator("navigator:browser");
        
        while (wndEnum.hasMoreElements())
            windows.push(wndEnum.getNext());
        
        return windows;
    },
    
    getTopBrowserWindow: function misc_getTopBrowserWindow() {
        return this.getTopWindowOfType("navigator:browser");
    },
    
    getTopWindowOfType: function misc_getTopWindowOfType(windowType) {
        let mediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
        return mediator.getMostRecentWindow(windowType);
    },
    
    hiddenWindows: {
        get appWindow() {
            let hiddenWindow;
            try {
                hiddenWindow = Cc["@mozilla.org/appshell/appShellService;1"]
                                   .getService(Ci.nsIAppShellService)
                                   .hiddenDOMWindow;
            } catch (e) {
                Cu.reportError(e);
            }
            
            if (!hiddenWindow)
                return null;
            
            delete this.appWindow;
            return (this.appWindow = hiddenWindow);
        },
        
        
        getFrame: function misc_getFrame(aFrameId, aFrameURL) {
            if (!aFrameURL || typeof aFrameURL != "string")
                throw new TypeError("aFrameURL must be a string.");
            
            let hiddenWindow = this.appWindow;
            if (!hiddenWindow)
                return null;
            
            let hiddenDoc = hiddenWindow.document;
            if (!hiddenDoc)
                return null;
            
            let url = aFrameURL;
            let id = aFrameId || btoa(url);
            let frameLoader = hiddenDoc.getElementById(id);
            
            if (!frameLoader) {
                frameLoader = hiddenDoc.createElement("iframe");
                frameLoader.setAttribute("id", id);
                frameLoader.setAttribute("src", url);
                hiddenDoc.documentElement.appendChild(frameLoader);
            }
            
            let contentWindow = frameLoader.contentWindow;
            
            if (contentWindow.location != url) {
                let thread = Cc["@mozilla.org/thread-manager;1"].getService().currentThread;
                let t = Date.now();
                while ((Date.now() - t) < 1000 && contentWindow.location != url)
                    thread.processNextEvent(true);
            }
            
            if (contentWindow.location != url) {
                Cu.reportError("Can't get hidden window for '" + aFrameURL + "'");
                return null;
            }
            
            return frameLoader;
        },
        
        
        removeFrame: function misc_removeFrame(aFrameId, aFrameURL) {
            if (!aFrameId && !aFrameURL)
                throw new TypeError("Need frame id or frame url.");
            
            let hiddenWindow = this.appWindow;
            if (!hiddenWindow)
                return;
            
            let hiddenDoc = hiddenWindow.document;
            if (!hiddenDoc)
                return;
            
            let id = aFrameId || btoa(aFrameURL);
            let frameLoader = hiddenDoc.getElementById(id);
            
            if (frameLoader)
                frameLoader.parentNode.removeChild(frameLoader);
        },
        
        
        getWindow: function misc_getWindow(aFrameURL) {
            let frameLoader = this.getFrame(null, aFrameURL);
            return frameLoader && frameLoader.contentWindow || null;
        }
    },
    openWindow: function misc_openWindow(parameters) {
        let window;
        
        if ("name" in parameters && parameters.name) {
            const WM = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
            if ((window = WM.getMostRecentWindow(parameters.name))) {
                window.focus();
                return window;
            }
        }
        
        let parent;
        let features = parameters.features || "";
        
        if (features.indexOf("__popup__") != -1) {
            let featuresHash = { __proto__: null };
            
            features.replace(/(^|,)__popup__($|,)/, "").split(",")
            .forEach(function(aFeatureString) {
                if (aFeatureString) {
                    let [name, value] = aFeatureString.split("=");
                    if (name)
                        featuresHash[name] = value;
                }
            });
            let addFeature = function(aFeatureString) {
                let [name, value] = aFeatureString.split("=");
                if (!(name in featuresHash))
                    featuresHash[name] = value;
            }
            
            addFeature("chrome");
            addFeature("dependent=yes");
            
            if (!Ci.nsIDOMGeoPositionAddress)
                addFeature("titlebar=no");
            
            if (sysutils.platformInfo.os.name != "windows")
                addFeature("popup=yes");
            let featuresMod = [];
            for (let [name, value] in Iterator(featuresHash))
                featuresMod.push(name + (value ? "=" + value : ""));
            
            features = featuresMod.join(",");
            if (!("parent" in parameters))
                parent = this.getTopBrowserWindow();
        }
        
        parent = parent || parameters.parent || null;
        
        const WW = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
        window = WW.openWindow(
            parent,
            parameters.url,
            parameters.name || "_blank",
            features,
            parameters.arguments || undefined
        );
        
        window.parameters = parameters;
        
        return window;
    },
    
    
    tryCreateFixupURI: function misc_tryCreateFixupURI(aString) {
        let URIFixup = Cc["@mozilla.org/docshell/urifixup;1"].getService(Ci.nsIURIFixup);
        try {
            return URIFixup.createFixupURI(aString, URIFixup.FIXUP_FLAG_NONE);
        } catch (e) {
            return null;
        }
    },
    
    
    mapKeysToArray: function misc_mapKeysToArray(map, filter) {
        return filter ? [key for (key in map) if (filter(key))]
                      : [key for (key in map)];
    },
    
    
    mapValsToArray: function misc_mapValsToArray(map, filter) {
        return filter ? [val for each (val in map) if (filter(key))]
                      : [val for each (val in map)];
    },
    
    
    separateItems: function misc_separateItems(input, check) {
        if (typeof check != "function")
            throw new CustomErrors.EArgType("check", "Function", check);
        
        let trueList = [];
        let falseList = [];
        
        for each (let item in input) {
            (check(item) ? trueList : falseList).push(item);
        }
        
        return [trueList, falseList];
    },
    get CryptoHash() {
        let CryptoHash = {
            getFromString: function CryptoHash_getFromString(aString, aAlgorithm) {
                let hash = this._createHash(aAlgorithm);
                let stream = strutils.utf8Converter.convertToInputStream(aString);
                this._updateHashFromStream(hash, stream);
                return this._binaryToHex(hash.finish(false));
            },
            _createHash: function CryptoHash__createHash(aAlgorithm) {
                let hash = Cc["@mozilla.org/security/hash;1"].createInstance(Ci.nsICryptoHash);
                hash.initWithString(aAlgorithm);
                return hash;
            },
            
            _binaryToHex: function CryptoHash__binaryToHex(aInput) {
                return [("0" + aInput.charCodeAt(i).toString(16)).slice(-2)
                            for (i in aInput)].join("");
            },
            
            _updateHashFromStream: function CryptoHash__updateHashFromStream(aHash, aStream) {
                let streamSize = aStream.available();
                if (streamSize)
                    aHash.updateFromStream(aStream, streamSize);
            }
        };
        
        delete this.CryptoHash;
        return (this.CryptoHash = CryptoHash);
    },
    
    
    parseLocale: function misc_parseLocale(localeString) {
        let components = localeString.match(this._localePattern);
        if (components)
            return {
                language: components[1],
                country: components[3],
                region: components[5]
            };
        return null;
    },
    
    
    findBestLocalizedValue: function misc_findBestLocalizedValue(map, forLocale) {
        const lpWeights = {
            language: 32,
            empty: 16,
            ru: 8,
            en: 4,
            country: 2,
            region: 1
        };
        
        let results = [];
        for (let key in map) {
            if (!map.hasOwnProperty(key)) continue;
            
            let weight = 0;
            if (key) {
                let locale = misc.parseLocale(key);
                
                for (let partName in lpWeights) {
                    if (partName in locale) {
                        let localePart = locale[partName];
                        if (partName == "language")
                            if (localePart in lpWeights)
                                weight += lpWeights[localePart];
                        if (localePart === forLocale[partName])
                            weight += lpWeights[partName];
                    }
                }
            } else {
                weight = lpWeights.empty;
            }
            
            results.push({
                key: key,
                weight: weight
            });
        }
        
        results.sort(function rule(a, b) {
            if (a.weight == b.weight)
                return 0;
            return a.weight > b.weight ? -1 : +1;
        });
        
        return results[0] && map[results[0].key];
    },
    
    get storageService() {
        let storageSvc = Cc["@mozilla.org/storage/service;1"].getService(Ci.mozIStorageService);
        delete this.storageService;
        return (this.storageService = storageSvc);
    },
    
    _localePattern: /^([a-z]{2})(-([A-Z]{2})(-(\w{2,5}))?)?$/
};



misc.Database = function misc_Database(storageFile, initStatement) {
    if (storageFile)
        this.open(storageFile, initStatement);
};

misc.Database.prototype = {
    constructor: misc.Database,
    
    
    open: function Database_open(storageFile, initStatement) {
        if (!(storageFile instanceof Ci.nsILocalFile))
            throw new TypeError("nsILocalFile required");
        
        this.close();
        
        try {
            this._connectAndInit(storageFile, initStatement);
        } catch(e) {
            Cu.reportError( strutils.formatError(e) );
            
            if (this._connection) {
                this._connection.close();
            }
            
            let journalFile = storageFile.parent;
            journalFile.append(storageFile.leafName + "-journal");
            fileutils.removeFileSafe(journalFile);
            
            switch (e.result) {
                case Cr.NS_ERROR_FILE_CORRUPTED:
                case Cr.NS_ERROR_STORAGE_IOERR:
                    Cu.reportError("misc.Database: will try to remove corrupt DB file");
                    storageFile.remove(false);
                    Cu.reportError("misc.Database: corrupt DB file removed");
            }
            
            this._connectAndInit(storageFile, initStatement);
        }
        this._storageFile = storageFile;
    },
    
    
    close: function Database_close(callback) {
        if (callback && (typeof callback != "function"))
            throw new CustomErrors.EArgType("callback", "Function", callback);
        
        if (this._connection) {
            if ("asyncClose" in this._connection)
                this._connection.asyncClose( callback ? {
                    complete: function Database_closeComplete() { callback(); }
                } : undefined);
            else {
                this._connection.close();
                if (callback)
                    callback();
            }
            this._connection = null;
        }
    },
    
    
    execQuery: function Database_execQuery(query, parameters) {
        let result = [];
        
        let statement = this._createStatement(query, parameters);
        try {
            let colNames = [];
            for (let i = statement.columnCount; i--;)
                colNames.push(statement.getColumnName(i));
            while (statement.executeStep()) {
                let row = {};
                for (let i = colNames.length; i--;) {
                    let colName = colNames[i];
                    row[colName] = statement.row[colName];
                }
                
                result.push(row);
            }
        }
        finally {
            statement.reset();
            statement.finalize();
        }
        
        return result;
    },
    
    
    execSimpleQuery: function Database_execSimpleQuery(query, parameters) {
        let firstResult = this.execQuery(query, parameters)[0];
        if (firstResult)
            for (let p in firstResult)
                if (firstResult.hasOwnProperty(p))
                    return firstResult[p];
        return;
    },
    
    
    execQueryAsync: function Database_execAsync(query, parameters, onCompletion) {
        if (onCompletion && (typeof onCompletion != "function"))
            throw new TypeError("Third argument must be a function.");
        
        let statement = this._createStatement(query, parameters, true);
        try {
            return statement.executeAsync( this._createStmtCallback(statement, onCompletion) );
        }
        finally {
            statement.reset();
            statement.finalize();
        }
    },
    
    
    get lastInsertRowID() {
        return (this._connection && this._connection.lastInsertRowID) || 0;
    },
    
    _connection: null,
    
    _connectAndInit: function Database__connectAndInit(storageFile, initStatement) {
        this._connection = misc.storageService.openDatabase(storageFile);
        if (!this._connection.connectionReady)
            throw new Error(this._connection.lastError);
        
        if (initStatement)
            this.execQuery(initStatement);
    },
    
    _createStmtCallback: function Database__createStmtCallback(statement, onCompletion) {
        let colNames = [];
        for (let i = 0, len = statement.columnCount; i < len; i++)
            colNames.push(statement.getColumnName(i));
        let numNames = colNames.length;
        
        return {
            _result: [],
            _error: undefined,
            
            handleResult: function Database_callback_handleResult(aResultSet) {
                let row;
                while ((row = aResultSet.getNextRow())) {
                    let resultRow = {};
                    for (let nameIdx = 0; nameIdx < numNames; nameIdx++) {
                        let colName = colNames[nameIdx];
                        resultRow[colName] = row.getResultByName(colName);
                    }
                    this._result.push(resultRow);
                }
            },
           
            handleError: function Database_callback_handleError(aError) {
                this._error = aError;
            },
           
            handleCompletion: function Database_callback_handleCompletion(aReason) {
                if (aReason == Ci.mozIStorageStatementCallback.REASON_CANCELED)
                    return;
                if (onCompletion)
                    onCompletion(this._result, this._error);
            }
        };
    },
    
    _createStatement: function Database__createStatement(query, parameters, asyncStatement) {
        if (!this._connection)
            throw new Error("Can't create statement. Database is closed.");
        
        let statement = this._connection.createStatement(query);
        
        
        if (parameters) {
            for (let name in parameters)
                statement.params[name] = parameters[name];
        }
        
        return statement;
    }
};
