"use strict";

const sysutils = {
    
    dump: function SysUtils_dump(what, depth ) {
        const brackets = {object: "{}", array: "[]"};
        
        function _dump(what, depth, stack) {
            if (!what)
                return "" + what;
            
            let currIndent = strutils.repeatString("  ", stack.length);
            stack.push(what);
            
            let useBr = sysutils.isArray(what) ? brackets.array : brackets.object;
            let res = useBr[0];
            for (let propName in what) {
                let valStr;
                try {
                    let val = what[propName];
                    if ((depth > stack.length)
                            && sysutils.isObject(val)
                        ) {
                        valStr = _dump(val, depth, stack);
                    }
                    else {
                        if (sysutils.isString(val))
                            valStr = ['"', val, '"'].join("");
                        else
                            valStr = String(val);
                    }
                }
                catch (e) {
                    valStr = "/conversion error/";
                }
                
                res += ["\n", currIndent, "  ", propName, ": ", valStr].join("");
            }
            res += ("\n" + currIndent + useBr[1]);
            
            stack.pop();
            
            return res;
        }
        
        return _dump(what, depth, []);
    },
    
    get versionComparator() {
        let comparator = Cc["@mozilla.org/xpcom/version-comparator;1"].getService(Ci.nsIVersionComparator);
        this.__defineGetter__("versionComparator", function getVersionComparator() comparator);
        return this.versionComparator;
    },
    
    get platformInfo() {
        let info = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo).QueryInterface(Ci.nsIXULRuntime);
        
        let os = info.OS;
        
        if (/^win/i.test(os))
            os = "windows";
        else if (/^linux/i.test(os))
            os = "linux";
        else if (/^darwin/i.test(os))
            os = "mac";
        
        let comparator = this.versionComparator;
        
        let simpleArchitecture = info.XPCOMABI.split("-")[0] || "x86";
        if (simpleArchitecture == "x86_64")
            simpleArchitecture = "x64";
        
        let platformInfo = {
            browser: {
                name: "firefox",
                version: {
                    toString: function PBVersion_toString() info.version,
                    isLessThan: function PBVersion_isLessThan(aVersion) comparator.compare(this, aVersion) < 0,
                    isGreaterThan: function PBVersion_isGreaterThan(aVersion) comparator.compare(this, aVersion) > 0,
                    isEqual: function PBVersion_isEqual(aVersion) comparator.compare(this, aVersion) == 0
                },
                architecture: info.XPCOMABI,
                simpleArchitecture: simpleArchitecture
            },
            
            os: {
                name: os
            }
        };
        
        this.__defineGetter__("platformInfo", function() platformInfo);
        return this.platformInfo;
    },
    
    isEmptyObject: function SysUtils_isEmptyObject(obj) {
        for (let n in obj)
            return false;
        return true;
    },
    
    objectsAreEqual: function SysUtils_objectsAreEqual(obj1, obj2) {
        if (typeof obj1 != "object" || typeof obj2 != "object")
            throw new TypeError("Objects expected");
        
        let props1 = [], 
            props2 = [];
        
        for (let propName in obj1)
            props1.push(propName);
        
        for (let propName in obj2)
            props2.push(propName);
        
        if (props1.length != props2.length)
            return false;
        
        let prop1, prop2;
        for each (let propName in props1) {
            if ( !(propName in obj2) )
                return false;
            
            prop1 = obj1[propName];
            prop2 = obj2[propName];
            
            if ( (typeof prop1 == "object") && (typeof prop2 == "object") ) {
                if ( !this.objectsAreEqual(prop1, prop2) )
                    return false;
            }
            else {
                if (prop1 !== prop2)
                    return false;
            }
        }
        
        return true;
    },

    freezeObj: (function SysUtils_freezeObj_factory() {
        
        const watcher = function SysUtils_freezeObj_watcher(key, oldValue, newValue) {
            throw new TypeError("Can not modify property [" + key + "]: " + this + " is frozen");
        }
        
        return function SysUtils_freezeObj(obj) {

            if (obj.Object_freeze_self) {
                obj.Object_freeze_self();
                return;
            }
            
            for (let key in obj) {
                if (obj.hasOwnProperty(key)) {
                    let value = obj[key];
                    if (value && typeof value === "object") {
                        sysutils.freezeObj(value);
                    }
                }
                
                obj.watch(key, watcher);
            }
            if (Object.seal) {
                Object.seal(obj);
            }
        }
    })(),

    
    copyObj: function SysUtils_copyObj(src, deep) {
        if (typeof src != "object" || !src)
            return src;
        
        if (this.isArray(src))
            return src.map(function(el) {return deep ? this.copyObj(el, deep) : el}, this);
        
        let result = {};
        
        for (let [name, value] in Iterator(src))
            result[name] = deep ? this.copyObj(value, deep) : value;
        
        return result;
    },
    
    
    mergeObj: function SysUtils_mergeObj() {
        let result = {};
        
        for (let i = 0; i < arguments.length; ++i) {
            let arg = arguments[i];
            this.copyProperties(arg, result);
        }
        
        return result;
    },
    
    
    copyProperties: function SysUtils_copyProperties(from, to, filter) {
        let filterFunc;
        if (filter) {
            filterFunc = (typeof filter == "function") ?
                filter :
                function SysUtils_copyFilter(propName) propName in filter;
        }
        let passValues = filterFunc && (filterFunc.length > 1);
        
        for (let propName in from) {
            if (from.hasOwnProperty(propName) &&
                    (!filterFunc || filterFunc(propName, passValues ? from[propName] : undefined)) )
                to[propName] = from[propName];
        }
    },
    isPlainObject: function SysUtils_isPlainObject(val) {
        return Object.getPrototypeOf(Object.getPrototypeOf(val)) === null;
    },
    isObject: function SysUtils_isObject(val) {
        return Boolean(val && Object.prototype.toString.call(val) == "[object Object]");
    },
    
    isArray: function SysUtils_isArray(val) {
        return Boolean(val && Object.prototype.toString.call(val) == "[object Array]");
    },
    
    isNumber: function SysUtils_isNumber(x) {
        return (typeof x == "number" && !isNaN(x));
    },
    
    isString: function SysUtils_isString(x) {
        return (typeof x == "string");
    },
    
    isBool: function SysUtils_isBool(x) {
        return (typeof x == "boolean");
    },
    cutNumberTo: function SysUtils_cutNumberTo(num, numDigits) {
        let intPart = Math.floor(num);
        let fracPart = num - intPart;
        let pow = Math.pow(10, numDigits);
        let cutFrac = Math.floor(fracPart * pow) / pow;
        return intPart + cutFrac;
    },
    
    defineLazyGetter: function SysUtils_defineLazyGetter(aObject, aName, aLambda) {
        aObject.__defineGetter__(aName, function SysUtils_lazyGetter() {
            delete aObject[aName];
            let getterValue = aLambda.apply(aObject);
            aObject.__defineGetter__(aName, function SysUtils_lazyGetter() {
                return getterValue;
            });
            return aObject[aName];
        });
    },
    
    defineLazyServiceGetter: function SysUtils_defineLazyServiceGetter(aObject, aName, aContract, aInterfaceName) {
        this.defineLazyGetter(aObject, aName, function SysUtils_lazyService() {
            return Cc[aContract].getService(Ci[aInterfaceName]);
        });
    },
    
    WeakObjectProxy: function WeakObjectProxy(object) {
        if (!sysutils.isObject(object))
            throw new CustomErrors.EArgType("object", "Object", object);
        
        let inst = (function weakRefInitializer() {
            let __storedObject = object;
            
            return {
                get isValid() {
                    return __storedObject !== undefined;
                },
                
                invoke: function WeakObjectProxy_invoke(methodName, methodArguments) {
                    if (!this.isValid)
                        throw new Error("This weak object is invalid");
                    let method = __storedObject[methodName];
                    if (typeof method !== "function")
                        throw new Error(strutils.formatString("No such method (%1)", [methodName]));
                    
                    let args = Array.slice(arguments, 1);
                    return method.apply(__storedObject, args);
                },
                
                __noSuchMethod__: function WeakObjectProxy___noSuchMethod__(methodName, methodArguments) {
                    return this.invoke.apply(this, [methodName].concat(methodArguments));
                },
                
                clear: function WeakObjectProxy_clear() {
                    __storedObject = undefined;
                }
            };
        })();
        
        return inst;
    },
    
    _Ci: Components.interfaces,
    _Cc: Components.classes
};
sysutils.Interface = function Interface(name, methods, properties) {
    this._name = name.toString();
    this._properties = {};
    
    if (methods !== undefined) {
        if (!sysutils.isArray(methods))
            throw new TypeError("Methods must be described by an array of strings");
        for each (let methodName in methods)
            this._properties[methodName] = "function";
    }
    
    if (properties !== undefined) {
        if (!sysutils.isPlainObject(properties))
            throw new TypeError("Properties must be described by a plain map of properties");
        sysutils.copyProperties(properties, this._properties);
    }
};

sysutils.Interface.prototype = {
    constructor: sysutils.Interface,
    
    get name() {
        return this._name;
    },
    
    
    isImplementedBy: function Interface_isImplementedBy(something, errors) {
        const errOut = sysutils.isArray(errors);
        
        for each (let [propName, propDescr] in Iterator(this._properties)) {
            if (!this._checkProperty(something, propName, propDescr)) {
                if (errOut)
                    errors.push(propName);
                return false;
            }
        }
        
        return true;
    },
    
    
    checkImplementation: function Interface_checkImplementation(something) {
        let missing = []
        if (!this.isImplementedBy(something, missing))
            throw new CustomErrors.ENoInterface(this._name, missing);
    },
    
    _name: undefined,
    _properties: null,
    
    _checkProperty: function Unterface__checkProperty(something, propName, propDescr) {
        if (!(propName in something))
            return false;
        
        if (propDescr === undefined)
            return true;
        
        let actualValue = something[propName];
        if (typeof propDescr == "string")
            return (typeof actualValue == propDescr);
        else if (typeof propDescr == "function")
            return (actualValue instanceof propDescr);
        else if (propDescr instanceof this.constructor)
            return (propDescr.isImplementedBy(actualValue));
        else
            return true;
    }
};


sysutils.Timer = function Timer(aCallback, aDelay, aRepeating, aMaxTimes) {
    this.callback = aCallback;
    
    this.repeatingDelay = (typeof aRepeating == "number" && aRepeating > 0) ? aRepeating : null;
    this.repeating = (typeof aRepeating == "boolean" || this.repeatingDelay) ? !!aRepeating : false;
    
    this.timesCounter = 0;
    this.maxTimes = (typeof aMaxTimes == "number" && aMaxTimes > 0) ? aMaxTimes : null;
    
    this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    let type = aRepeating ? this.timer.TYPE_REPEATING_SLACK :  this.timer.TYPE_ONE_SHOT;
    this.timer.initWithCallback(this, aDelay, type);
    
    Cc["@mozilla.org/observer-service;1"]
        .getService(Ci.nsIObserverService)
        .addObserver(this, "xpcom-shutdown", false);
};

sysutils.Timer.prototype = {
    observe: function Timer_observe(aSubject, aTopic, aData) {
        if (aTopic === "xpcom-shutdown")
            this.cancel();
    },
    
    get isRunning() {
        return !!this.timer;
    },
    
    cancel: function Timer_cancel() {
        if (!this.timer)
            return;
        
        this.timer.cancel();
        this.timer = null;
        this.callback = null;
        
        Cc["@mozilla.org/observer-service;1"]
            .getService(Ci.nsIObserverService)
            .removeObserver(this, "xpcom-shutdown", false);
    },
    
    notify: function Timer_notify(aTimer) {
        let result = this.callback();
        
        this.timesCounter++;
        
        if (this.timesCounter == 1 && this.repeatingDelay)
            this.delay = this.repeatingDelay;
        
        if (!this.repeating || (this.maxTimes && this.timesCounter >= this.maxTimes))
            this.cancel();
        
        return result;
    },
    
    set delay(val) {
        if (this.timer)
            this.timer.delay = val;
    },
    
    QueryInterface: function Timer_QueryInterface(iid) {
        if (iid.equals(Ci.nsISupports) ||
            iid.equals(Ci.nsITimerCallback) ||
            iid.equals(Ci.nsIObserver))
            return this;
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
};
