"use strict";

const EXPORTED_SYMBOLS = ["prefRoot"];

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;

Cu.import("resource://yasearch-mod/Preferences.jsm");
Cu.import("resource://yasearch-mod/Log4Moz.jsm");
Cu.import("resource://yasearch-mod/Foundation.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

let log = new function(){
    var logger = Log4Moz.repository.getLogger("yasearch.modules.prefRoot");
    this.error = function log_error(error) {
        logger.error(error);
        logger.debug(error.stack);
    };
};

function PrefNode (args) {
    this.branch = args.branch;
    this.site = args.site;
    
    this.prefs = new Preferences(args);
}

PrefNode.prototype = new function(){

    this.put = FFallPipe(put, FValue(false));
    this.get = FFallPipe(get, getDefault, FValue(null));
    this.exists = Delegator("has");
    this.changed = Delegator("isSet");
    this.reset = Delegator("resetBranch");
    this.lock = Delegator("lock");
    this.unlock = Delegator("unlock");
    this.locked = Delegator("locked");
    this.listen = Delegator("observe");
    this.forget = Delegator("ignore");
    
    this.append = function PrefNode_append(leaf) {
        return new PrefNode({site: this.prefs._site, branch: this.prefs._prefBranch + leaf});
    }
    
    this.site = function PrefNode_site(site) {
        return new PrefNode({site: site, branch: this.prefs._prefBranch});
    }
    
    this.createObserver = function PrefNode_createObserver(eventName, handler) {
        return new PrefObserver({ eventName: eventName, handler: handler, prefNode: this }); 
    }
    
    this.toString = function PrefNode_toString() {
        return JSON.stringify(this.branch) + " = " + JSON.stringify(this.get(""));
    }
    
    function get(defaultValue) {
        let prefValue = this.prefs.get("", null);
        
        if (prefValue === null) {
            return getDefault.apply(this, arguments);
        }
        
        let constructor = constructorOf(defaultValue);
        if (constructor) {
            prefValue = new constructor(prefValue);
            prefValue = prefValue.valueOf();
        }

        return prefValue;
    }
    
    function getDefault(defaultValue) {
        if (!defaultValue) {
            return defaultValue;
        }
        if (typeOf(defaultValue) === "date") {
            return defaultValue;
        }
        return defaultValue.valueOf();
    }
    
    function put(prefValue) {
        prefValue = complex2primitive(prefValue);
        if (prefValue === void 0) {
            return false;
        }
        
        if (prefValue === null) {
            this.prefs.reset("");
        } else {
            this.prefs.overwrite("", prefValue);
        }
        
        return true;
    }
    
    function Delegator(methodName) {
        return function Delegator_instance() {
            var method = this.prefs[methodName];
            Array.prototype.unshift.call(arguments, "");
            return method.apply(this.prefs, arguments);
        }
    }
        
}

const prefService =
    Cc["@mozilla.org/preferences-service;1"]
    .getService(Ci.nsIPrefService)
    .QueryInterface(Ci.nsIPrefBranch2);

const PrefObserver = function ({ eventName, handler, prefNode }) {
    this.eventName = eventName;
    this.prefNode = prefNode;
    this.handler = handler;
    this.observe = (eventName === "*") ? this.observeAll : this.observeTopic;
}
PrefObserver.prototype = new function () {
    this.QueryInterface= XPCOMUtils.generateQI([Ci.nsIObserver]);
    
    this.listen = function PrefObserver_listen() {
        prefService.addObserver(this.prefNode.branch, this, false);
        return this;
    }
    
    this.forget = function PrefObserver_forget() {
        prefService.removeObserver(this.prefNode.branch, this);
        return this;
    }
    
    this.observeAll = function PrefObserver_observeAll(aSubject, aTopic, aData) {
        return this.handler.apply(this, arguments);
    }
    
    this.observeTopic = function PrefObserver_observeTopic(aSubject, aTopic, aData) {
        if (aTopic !== this.eventName) return null;
        return this.handler.apply(this, arguments);
    }
}

let prefRoot = new PrefNode("");

function FFallPipe() {
    var pipe = arguments;
    return function FallPipe_instance() {
        for (let i = 0; i < pipe.length; ++i) {
            try {
                return pipe[i].apply(this, arguments);
            } catch (error) {
                log.error(error);
            }
        }
    }
}

function FArg(index) {
    return function FArg_instance() {
        return arguments[index];
    }
}

function FValue(value) {
    return function FValue_instance() {
        return value;
    }
}

function constructorOf(value) {
    let type = typeOf(value);
    switch (type) {
        case "date":
            return function(val) {
                let date = new Date(Number(val));
                if (isNaN(date.getTime())) {
                    throw new Error("Wrong timestamp [" + val + "]");
                }
                this.valueOf = FValue(date);
            }
        case "boolean":
            return function(val) {
                this.valueOf = FValue(Boolean(val));
            }
        case "object":
        case "function":
            return value && value.fromJSON || function JSON_constructor() JSON.parse.apply(JSON, arguments);
        case "undefined":
        case "null":
            return value;
        default:
            return value.constructor;
    }
    return value;
}

function complex2primitive(value) {
    let type = typeOf(value);
    switch (type) {
        case "date":
            return String(value.getTime());
        case "object":
        case "function":
            if (!value) {
                return value;
            }
            if (value && value.toJSON) {
                return value.toJSON();
            }
            return JSON.stringify(value);
        default:
            return value;
    }
}

let Object_toString = Object.prototype.toString;

function typeOf(value) {
    let type = typeof value;
    switch (type) {
        case "object":
            if (value === null) {
                return "null";
            }
            if (Object_toString.call(value) === "[object Date]") {
                return "date";
            }
        default:
            return type;
    }
}

